"""The parsing module contains the code for the parsing of command-line arguments."""
from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path

from chemex import __version__, chemex
from chemex.parameters.spin_system import SpinSystem
from chemex.tools.pick_cest import pick_cest
from chemex.tools.plot_param import plot_param


def build_parser():
    """Parse the command-line arguments."""
    description = (
        "ChemEx is an analysis program for chemical exchange detected by "
        "NMR. It is designed to take almost any kind of NMR data to aid the "
        "analysis, but the principle techniques are CPMG relaxation "
        "dispersion and Chemical Exchange Saturation Transfer."
    )

    parser = ArgumentParser(description=description, prog="chemex")

    parser.set_defaults(func=lambda _: parser.print_usage())

    parser.add_argument(
        "--version", action="version", version=f"{parser.prog} {__version__}"
    )

    subparsers = parser.add_subparsers(dest="commands")

    # parser for the positional argument "fit"
    fit_parser = subparsers.add_parser("fit", help="Start a fit")

    fit_parser.set_defaults(func=chemex.run)

    fit_parser.add_argument(
        "-e",
        "--experiments",
        type=Path,
        metavar="FILE",
        nargs="+",
        required=True,
        help="Input file(s) containing experimental setup and data location",
    )

    fit_parser.add_argument(
        "-p",
        "--parameters",
        type=Path,
        metavar="FILE",
        nargs="+",
        required=True,
        help="Input file(s) containing the initial values of fitting parameters",
    )

    fit_parser.add_argument(
        "-m",
        "--method",
        type=Path,
        metavar="FILE",
        nargs="+",
        help="Input file(s) containing the fitting method",
    )

    fit_parser.add_argument(
        "-o",
        "--output",
        type=Path,
        metavar="DIR",
        default="./Output",
        help="Directory for output files (default: './Output')",
    )

    fit_parser.add_argument(
        "-d",
        "--model",
        metavar="MODEL",
        default="2st",
        help="Exchange model used to fit the data (default: '2st')",
    )

    fit_parser.add_argument(
        "--plot",
        type=str,
        choices=["nothing", "normal", "all"],
        default="normal",
        help="Plotting level (default: 'normal')",
    )

    fit_parser.add_argument(
        "--include",
        dest="include",
        metavar="ID",
        nargs="+",
        help="Residue(s) to include in the fit",
        type=SpinSystem,
    )

    fit_parser.add_argument(
        "--exclude",
        dest="exclude",
        metavar="ID",
        nargs="+",
        help="Residue(s) to exclude from the fit",
        type=SpinSystem,
    )

    # parser for the positional argument "simulate"
    simulate_parser = subparsers.add_parser("simulate", help="Start a simulation")

    simulate_parser.set_defaults(func=chemex.run)

    simulate_parser.add_argument(
        "-e",
        "--experiments",
        type=Path,
        metavar="FILE",
        nargs="+",
        required=True,
        help="Input file(s) containing experimental setup and data location",
    )

    simulate_parser.add_argument(
        "-p",
        "--parameters",
        type=Path,
        metavar="FILE",
        nargs="+",
        required=True,
        help="Input file(s) containing the values of parameters",
    )

    simulate_parser.add_argument(
        "-o",
        "--output",
        type=Path,
        metavar="DIR",
        default="./OutputSim",
        help="Directory for output files (default: './OutputSim')",
    )

    simulate_parser.add_argument(
        "-d",
        "--model",
        metavar="MODEL",
        default="2st",
        help="Exchange model used to simulate the data (default: '2st')",
    )

    simulate_parser.add_argument(
        "--plot",
        type=str,
        choices=["nothing", "normal"],
        default="normal",
        help="Plotting level (default: 'normal')",
    )

    simulate_parser.add_argument(
        "--include",
        dest="include",
        metavar="ID",
        nargs="+",
        help="Residue(s) to include in the simulation",
        type=SpinSystem,
    )

    simulate_parser.add_argument(
        "--exclude",
        dest="exclude",
        metavar="ID",
        nargs="+",
        help="Residue(s) to exclude from the simulation",
        type=SpinSystem,
    )

    # parser for the positional argument "pick_cest"
    pick_cest_parser = subparsers.add_parser(
        "pick_cest", help="Plot CEST profiles for dip picking"
    )

    pick_cest_parser.set_defaults(func=pick_cest)

    pick_cest_parser.add_argument(
        "-e",
        "--experiments",
        type=Path,
        metavar="FILE",
        nargs="+",
        required=True,
        help="Input file(s) containing experimental setup and data location",
    )

    pick_cest_parser.add_argument(
        "-o",
        "--output",
        type=Path,
        metavar="DIR",
        default="./Sandbox",
        help="Directory for output files (default: './Sandbox')",
    )

    # parser for the positional argument "pick_cest"
    plot_param_parser = subparsers.add_parser(
        "plot_param", help="Plot one selected parameter from a 'parameters.fit' file"
    )

    plot_param_parser.set_defaults(func=plot_param)

    plot_param_parser.add_argument(
        "-p",
        "--parameters",
        type=Path,
        metavar="FILE",
        nargs="+",
        required=True,
        help="Input file containing the fitted parameters to be plotted",
    )

    plot_param_parser.add_argument(
        "-n",
        "--parname",
        metavar="NAME",
        required=True,
        help="Name of the parameter to plot",
    )

    return parser



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Annotated

from annotated_types import Len
from pydantic import AfterValidator, BeforeValidator, RootModel

from chemex.configuration.base import ensure_list
from chemex.parameters.name import ParamName
from chemex.toml import read_toml_multi

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path


def rename_section(section_name: str) -> str:
    if section_name == "global":
        return ""
    return f"{section_name},nuc->"


ValuesType = Annotated[list[float], Len(max_length=4), BeforeValidator(ensure_list)]
LowerCaseString = Annotated[str, BeforeValidator(str.lower)]
ValuesDictType = dict[LowerCaseString, ValuesType]
SectionType = Annotated[LowerCaseString, AfterValidator(rename_section)]
ParamsConfigType = dict[SectionType, ValuesDictType]
ParamsConfigModel = RootModel[ParamsConfigType]


@dataclass(frozen=True)
class DefaultSetting:
    value: float
    min: float | None = None
    max: float | None = None
    brute_step: float | None = None


DefaultType = tuple[ParamName, DefaultSetting]
DefaultListType = list[DefaultType]


def build_default_list(params_config: ParamsConfigModel) -> DefaultListType:
    defaults: DefaultListType = []
    for section, params in params_config.root.items():
        for key, values in params.items():
            pname = ParamName.from_section(f"{section}{key}")
            default_values = DefaultSetting(*values)
            defaults.append((pname, default_values))
    return defaults


def read_defaults(filenames: Iterable[Path]) -> DefaultListType:
    config = read_toml_multi(filenames)
    param_config = ParamsConfigModel.model_validate(config)
    return build_default_list(param_config)



"""Factories for creating parameter settings."""
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

if TYPE_CHECKING:
    from collections.abc import Callable

    from chemex.configuration.conditions import Conditions
    from chemex.parameters.setting import ParamLocalSetting

    SettingsType = dict[str, ParamLocalSetting]
    SettingMakerType = Callable[[Conditions], SettingsType]


class Factory:
    """Factory for creating all parts of an experiment."""

    setting_makers_registry: ClassVar[dict[str, SettingMakerType]] = {}

    def register(self, name: str, setting_maker: SettingMakerType):
        """Register a new setting maker."""
        self.setting_makers_registry[name] = setting_maker

    def create(self, name: str, conditions: Conditions) -> SettingsType:
        try:
            return self.setting_makers_registry[name](conditions)
        except KeyError:
            msg = f"Unknown type {name!r}"
            raise ValueError(msg) from None

    @property
    def set(self) -> set[str]:
        return set(self.setting_makers_registry)


model_factory = Factory()



"""The chemex module provides the entry point for the chemex script."""
from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from chemex.cli import build_parser
from chemex.configuration.methods import Method, Selection, read_methods
from chemex.configuration.parameters import read_defaults
from chemex.experiments.builder import build_experiments
from chemex.experiments.loader import register_experiments
from chemex.messages import (
    print_logo,
    print_no_data,
    print_reading_defaults,
    print_reading_methods,
    print_running_simulations,
    print_start_fit,
)
from chemex.models import model
from chemex.models.loader import register_kinetic_settings
from chemex.optimize.fitting import run_methods
from chemex.optimize.helper import execute_simulation
from chemex.parameters import database

if TYPE_CHECKING:
    from argparse import Namespace

    from chemex.containers.experiments import Experiments


def run_fit(args: Namespace, experiments: Experiments):
    # Filter datapoints out if necessary (e.g., on-resonance filter CEST)
    experiments.filter()

    if args.method is not None:
        print_reading_methods()
        methods = read_methods(args.method)
    else:
        methods = {"": Method()}

    print_start_fit()
    run_methods(experiments, methods, args.output, args.plot)


def run_sim(args: Namespace, experiments: Experiments):
    print_running_simulations()

    path = args.output
    plot = args.plot == "normal"

    database.fix_all_parameters()
    experiments.back_calculate()

    execute_simulation(experiments, path, plot=plot)


def run(args: Namespace):
    """Run the fit or simulation."""
    # Parse kinetics model
    model.set_model(args.model)

    # Read experimental setup and data
    selection = Selection(args.include, args.exclude)
    experiments = build_experiments(args.experiments, selection)

    if not experiments:
        print_no_data()
        sys.exit()

    # Read initial values of fitting/fixed parameters
    print_reading_defaults()
    defaults = read_defaults(args.parameters)
    database.set_param_defaults(defaults)

    if args.commands == "simulate":
        run_sim(args, experiments)
    else:
        run_fit(args, experiments)


def main():
    """Do all the magic."""
    print_logo()

    register_kinetic_settings()
    register_experiments()

    parser = build_parser()
    args = parser.parse_args()
    args.func(args)



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np
from scipy.optimize import root

from chemex.models.constraints import pop_3st
from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting
from chemex.parameters.userfunctions import user_function_registry

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions
    from chemex.typing import ArrayFloat

NAME = "3st_monomer_dimer_tetramer"

TP = ("temperature", "p_total")


def calculate_residuals(
    concentrations: ArrayFloat,
    p_total: float,
    kd1: float,
    kd2: float,
) -> ArrayFloat:
    monomer, dimer, tetramer = concentrations
    return np.array(
        [
            monomer + 2.0 * dimer + 4.0 * tetramer - p_total,
            kd1 * dimer - monomer**2,
            kd2 * tetramer - dimer**2,
        ],
    )


@lru_cache(maxsize=100)
def calculate_concentrations(
    p_total: float,
    kd1: float,
    kd2: float,
) -> dict[str, float]:
    concentrations_start = (p_total, 0.0, 0.0)
    results = root(calculate_residuals, concentrations_start, args=(p_total, kd1, kd2))
    return {
        "monomer": results["x"][0],
        "dimer": results["x"][1],
        "tetramer": results["x"][2],
    }


def make_settings_3st_monomer_dimer_tetramer(
    conditions: Conditions,
) -> dict[str, ParamLocalSetting]:
    p_total = conditions.p_total
    if p_total is None:
        msg = f"'p_total' must be specified to use the '{NAME}' model"
        raise ValueError(msg)
    return {
        "kd1": ParamLocalSetting(
            name_setting=NameSetting("kd1", "", ("temperature",)),
            value=1e-6,
            min=0.0,
            vary=True,
        ),
        "kd2": ParamLocalSetting(
            name_setting=NameSetting("kd2", "", ("temperature",)),
            value=1e-6,
            min=0.0,
            vary=True,
        ),
        "koff1": ParamLocalSetting(
            name_setting=NameSetting("koff1", "", ("temperature",)),
            value=100.0,
            min=0.0,
            vary=True,
        ),
        "koff2": ParamLocalSetting(
            name_setting=NameSetting("koff2", "", ("temperature",)),
            value=100.0,
            min=0.0,
            vary=True,
        ),
        "kon1": ParamLocalSetting(
            name_setting=NameSetting("kon1", "", ("temperature",)),
            min=0.0,
            expr="{koff1} / max({kd1}, 1e-32)",
        ),
        "kon2": ParamLocalSetting(
            name_setting=NameSetting("kon2", "", ("temperature",)),
            min=0.0,
            expr="{koff2} / max({kd2}, 1e-32)",
        ),
        "c_monomer": ParamLocalSetting(
            name_setting=NameSetting("monomer", "", TP),
            expr=f"concentrations({p_total}, {{kd1}}, {{kd2}})['monomer']",
        ),
        "c_dimer": ParamLocalSetting(
            name_setting=NameSetting("dimer", "", TP),
            expr=f"concentrations({p_total}, {{kd1}}, {{kd2}})['dimer']",
        ),
        "kab": ParamLocalSetting(
            name_setting=NameSetting("kab", "", TP),
            expr="2.0 * {kon1} * {monomer}",
        ),
        "kba": ParamLocalSetting(
            name_setting=NameSetting("kba", "", TP),
            expr="{koff1}",
        ),
        "kbc": ParamLocalSetting(
            name_setting=NameSetting("kab", "", TP),
            expr="2.0 * {kon2} * {dimer}",
        ),
        "kcb": ParamLocalSetting(
            name_setting=NameSetting("kba", "", TP),
            expr="{koff2}",
        ),
        "pa": ParamLocalSetting(
            name_setting=NameSetting("pa", "", TP),
            expr="pop_3st({kab}, {kba}, 0.0, 0.0, {kbc}, {kcb})['pa']",
        ),
        "pb": ParamLocalSetting(
            name_setting=NameSetting("pb", "", TP),
            expr="pop_3st({kab}, {kba}, 0.0, 0.0, {kbc}, {kcb})['pb']",
        ),
        "pc": ParamLocalSetting(
            name_setting=NameSetting("pc", "", TP),
            expr="pop_3st({kab}, {kba}, 0.0, 0.0, {kbc}, {kcb})['pc']",
        ),
    }


def register() -> None:
    model_factory.register(
        name=NAME,
        setting_maker=make_settings_3st_monomer_dimer_tetramer,
    )
    user_functions = {
        "concentrations": calculate_concentrations,
        "pop_3st": pop_3st,
    }
    user_function_registry.register(name=NAME, user_functions=user_functions)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal

import numpy as np
from numpy.typing import NDArray

from chemex.configuration.data import RelaxationDataSettings
from chemex.configuration.experiment import (
    ExperimentConfig,
    RelaxationSettings,
    ToBeFitted,
)
from chemex.containers.dataset import load_exsy_dataset
from chemex.experiments.factories import Creators, factories
from chemex.filterers import PlanesFilterer
from chemex.nmr.basis import Basis
from chemex.nmr.liouvillian import LiouvillianIS
from chemex.nmr.spectrometer import Spectrometer
from chemex.plotters.exsy import EXSYPlotter
from chemex.printers.data import EXSYPrinter

if TYPE_CHECKING:
    from chemex.containers.data import Data
    from chemex.parameters.spin_system import SpinSystem

# Type definitions
NDArrayFloat = NDArray[np.float_]
NDArrayBool = NDArray[np.bool_]


EXPERIMENT_NAME = "noesyfpgpph19"


class Noesyfpgpph19Settings(RelaxationSettings):
    name: Literal["noesyfpgpph19"]


class Noesyfpgpph19Config(
    ExperimentConfig[Noesyfpgpph19Settings, RelaxationDataSettings],
):
    @property
    def to_be_fitted(self) -> ToBeFitted:
        return ToBeFitted(rates=["r1_i"], model_free=["tauc", "s2"])


def build_spectrometer(
    config: Noesyfpgpph19Config,
    spin_system: SpinSystem,
) -> Spectrometer:
    conditions = config.conditions

    basis = Basis(type="iz", spin_system="hn")
    liouvillian = LiouvillianIS(spin_system, basis, conditions)

    return Spectrometer(liouvillian)


@dataclass
class Noesyfpgpph19Sequence:
    settings: Noesyfpgpph19Settings

    def calculate(self, spectrometer: Spectrometer, data: Data) -> np.ndarray:
        times = data.metadata["times"]
        states1 = data.metadata["states1"]
        states2 = data.metadata["states2"]

        # Getting the starting magnetization
        equilibrium = spectrometer.get_equilibrium()

        # Calculate delay propagators
        delays = spectrometer.delays(times)

        intensities = []
        for state1, state2, delay in zip(states1, states2, delays):
            start = spectrometer.keep(equilibrium, [f"iz_{state1}"])
            spectrometer.detection = f"[iz_{state2}]"
            intensities.append(spectrometer.detect(delay @ start))

        return np.array(intensities)

    def is_reference(self, metadata: NDArrayFloat) -> NDArrayBool:
        return np.full_like(metadata, False, dtype=np.bool_)


def register() -> None:
    creators = Creators(
        config_creator=Noesyfpgpph19Config,
        spectrometer_creator=build_spectrometer,
        sequence_creator=Noesyfpgpph19Sequence,
        dataset_creator=load_exsy_dataset,
        filterer_creator=PlanesFilterer,
        printer_creator=EXSYPrinter,
        plotter_creator=EXSYPlotter,
    )
    factories.register(type=EXPERIMENT_NAME, creators=creators)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal

import numpy as np

from chemex.configuration.data import ShiftDataSettings
from chemex.configuration.experiment import ExperimentConfig, ShiftSettings, ToBeFitted
from chemex.containers.dataset import load_shift_dataset
from chemex.experiments.factories import Creators, factories
from chemex.filterers import NoFilterer
from chemex.nmr.basis import Basis
from chemex.nmr.liouvillian import LiouvillianIS
from chemex.nmr.spectrometer import Spectrometer
from chemex.plotters.shift import ShiftPlotter
from chemex.printers.data import ShiftPrinter

if TYPE_CHECKING:
    from chemex.containers.data import Data
    from chemex.parameters.spin_system import SpinSystem
    from chemex.typing import ArrayBool, ArrayFloat


EXPERIMENT_NAME = "shift_15n_sq"


class Shift15NSqSettings(ShiftSettings):
    name: Literal["shift_15n_sq"]
    observed_state: Literal["a", "b", "c", "d"] = "a"

    @property
    def cs_i_name(self) -> str:
        return f"cs_i_{self.observed_state}"


class Shift15NSqConfig(ExperimentConfig[Shift15NSqSettings, ShiftDataSettings]):
    @property
    def to_be_fitted(self) -> ToBeFitted:
        state = self.experiment.observed_state
        return ToBeFitted(
            rates=[f"cs_i_{state}"],
            model_free=[f"cs_i_{state}"],
        )


def build_spectrometer(
    config: Shift15NSqConfig,
    spin_system: SpinSystem,
) -> Spectrometer:
    conditions = config.conditions

    basis = Basis(type="ixy", spin_system="nh")
    liouvillian = LiouvillianIS(spin_system, basis, conditions)
    return Spectrometer(liouvillian)


def _find_nearest(array: ArrayFloat, value: float) -> float:
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx]


@dataclass
class Shift15NSqSequence:
    settings: Shift15NSqSettings

    def calculate(self, spectrometer: Spectrometer, _data: Data) -> ArrayFloat:
        ppm_i = spectrometer.liouvillian.ppm_i
        ref_shift_i = spectrometer.par_values[self.settings.cs_i_name] * ppm_i
        shifts = spectrometer.calculate_shifts()
        shift_sq = _find_nearest(shifts, ref_shift_i)
        return np.array([shift_sq / ppm_i])

    def is_reference(self, metadata: ArrayFloat) -> ArrayBool:
        return np.full_like(metadata, False, dtype=np.bool_)


def register() -> None:
    creators = Creators(
        config_creator=Shift15NSqConfig,
        spectrometer_creator=build_spectrometer,
        sequence_creator=Shift15NSqSequence,
        dataset_creator=load_shift_dataset,
        filterer_creator=NoFilterer,
        printer_creator=ShiftPrinter,
        plotter_creator=ShiftPlotter,
    )
    factories.register(type=EXPERIMENT_NAME, creators=creators)



from __future__ import annotations

from pathlib import Path
from typing import Annotated, Any, Literal

from pydantic import BeforeValidator, field_validator

from chemex.configuration.base import BaseModelLowerCase, ensure_list
from chemex.parameters.spin_system import PydanticSpinSystem


class DataSettings(BaseModelLowerCase):
    path: Path = Path("./")
    scaled: bool = True


PathList = Annotated[list[Path], BeforeValidator(ensure_list)]


class RelaxationDataSettings(DataSettings):
    error: Literal["file", "duplicates"] = "file"
    filter_planes: list[int] = []
    profiles: dict[PydanticSpinSystem, PathList] = {}

    @field_validator("error", mode="before")
    def to_lower(cls, error: Any) -> str:
        if isinstance(error, str):
            return error.lower()
        return error


class CestDataSettings(DataSettings):
    error: Literal["file", "duplicates", "scatter"] = "file"
    filter_planes: list[int] = []
    filter_offsets: list[tuple[float, float]] = [(0.0, 0.0)]
    filter_ref_planes: bool = False
    profiles: dict[PydanticSpinSystem, PathList] = {}

    @field_validator("error", mode="before")
    def to_lower(cls, error: Any) -> str:
        if isinstance(error, str):
            return error.lower()
        return error


class CestDataSettingsNoRef(CestDataSettings):
    filter_ref_planes: bool = True


class ShiftDataSettings(DataSettings):
    error: Literal["file", "duplicates"] = "file"
    scaled: bool = False

    @field_validator("error", mode="before")
    def to_lower(cls, error: Any) -> str:
        if isinstance(error, str):
            return error.lower()
        return error



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np
from scipy.optimize import root

from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting
from chemex.parameters.userfunctions import user_function_registry

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions
    from chemex.typing import ArrayFloat


NAME = "4st_binding_3_bound_states"

TPL = ("temperature", "p_total", "l_total")


def calculate_residuals(
    concentrations: ArrayFloat,
    p_total: float,
    l_total: float,
    kd_ab: float,
    keq_bc: float,
    keq_cd: float,
) -> ArrayFloat:
    p_, l_, pl1, pl2, pl3 = concentrations
    return np.array(
        [
            l_total - (l_ + pl1 + pl2 + pl3),
            p_total - (p_ + pl1 + pl2 + pl3),
            kd_ab * pl1 - p_ * l_,
            keq_bc * pl1 - pl2,
            keq_cd * pl2 - pl3,
        ],
    )


@lru_cache(maxsize=100)
def calculate_concentrations(
    p_total: float,
    l_total: float,
    kd_ab: float,
    keq_bc: float,
    keq_cd: float,
) -> dict[str, float]:
    concentrations_start = (p_total, l_total, 0.0, 0.0, 0.0)
    results = root(
        calculate_residuals,
        concentrations_start,
        args=(p_total, l_total, kd_ab, keq_bc, keq_cd),
    )
    p_, l_, pl1, pl2, pl3 = results["x"]
    return {"p": p_, "l": l_, "pl1": pl1, "pl2": pl2, "pl3": pl3}


def make_settings_4st_binding_3_bound_states(
    conditions: Conditions,
) -> dict[str, ParamLocalSetting]:
    p_total = conditions.p_total
    l_total = conditions.l_total
    if p_total is None:
        msg = f"'p_total' must be specified to use the '{NAME}' model"
        raise ValueError(msg)
    if l_total is None:
        msg = f"'l_total' must be specified to use the '{NAME}' model"
        raise ValueError(msg)
    return {
        "koff_ab": ParamLocalSetting(
            name_setting=NameSetting("koff_ab", "", ("temperature",)),
            value=100.0,
            min=0.0,
            max=1e6,
            vary=True,
        ),
        "kd_ab": ParamLocalSetting(
            name_setting=NameSetting("kd_ab", "", ("temperature",)),
            value=1e-6,
            min=0.0,
            max=1.0,
            vary=True,
        ),
        "kex_bc": ParamLocalSetting(
            name_setting=NameSetting("kex_bc", "", ("temperature",)),
            value=1000.0,
            min=0.0,
            max=1e6,
            vary=True,
        ),
        "keq_bc": ParamLocalSetting(
            name_setting=NameSetting("keq_bc", "", ("temperature",)),
            value=1.0,
            min=0.0,
            max=100.0,
            vary=True,
        ),
        "kex_cd": ParamLocalSetting(
            name_setting=NameSetting("kex_cd", "", ("temperature",)),
            value=1000.0,
            min=0.0,
            max=1e6,
            vary=True,
        ),
        "keq_cd": ParamLocalSetting(
            name_setting=NameSetting("keq_cd", "", ("temperature",)),
            value=1.0,
            min=0.0,
            max=100.0,
            vary=True,
        ),
        "kon_ab": ParamLocalSetting(
            name_setting=NameSetting("kon_ab", "", ("temperature",)),
            expr="{koff_ab} / max({kd_ab}, 1e-32)",
        ),
        "c_p": ParamLocalSetting(
            name_setting=NameSetting("c_p", "", TPL),
            expr=f"concentrations({p_total},{l_total},{{kd_ab}},{{keq_bc}},{{keq_cd}})['p']",
        ),
        "c_l": ParamLocalSetting(
            name_setting=NameSetting("c_l", "", TPL),
            expr=f"concentrations({p_total},{l_total},{{kd_ab}},{{keq_bc}},{{keq_cd}})['l']",
        ),
        "c_pl1": ParamLocalSetting(
            name_setting=NameSetting("c_pl1", "", TPL),
            expr=f"concentrations({p_total},{l_total},{{kd_ab}},{{keq_bc}},{{keq_cd}})['pl1']",
        ),
        "c_pl2": ParamLocalSetting(
            name_setting=NameSetting("c_pl2", "", TPL),
            expr=f"concentrations({p_total},{l_total},{{kd_ab}},{{keq_bc}},{{keq_cd}})['pl2']",
        ),
        "c_pl3": ParamLocalSetting(
            name_setting=NameSetting("c_pl3", "", TPL),
            expr=f"concentrations({p_total},{l_total},{{kd_ab}},{{keq_bc}},{{keq_cd}})['pl3']",
        ),
        "c_pl": ParamLocalSetting(
            name_setting=NameSetting("c_pl", "", TPL),
            expr="{c_pl1}+{c_pl2}+{c_pl3}",
        ),
        "kd_eff": ParamLocalSetting(
            name_setting=NameSetting("kd_eff", "", TPL),
            expr="({c_p}+{c_l}) / max({c_pl}, 1e-32)",
        ),
        "kab": ParamLocalSetting(
            name_setting=NameSetting("kab", "", TPL),
            expr="{kon_ab} * {c_l}",
        ),
        "kba": ParamLocalSetting(
            name_setting=NameSetting("kba", "", TPL),
            expr="{koff_ab}",
        ),
        "kbc": ParamLocalSetting(
            name_setting=NameSetting("kbc", "", TPL),
            expr="{keq_bc}*{kex_bc}/(1.0+{keq_bc})",
        ),
        "kcb": ParamLocalSetting(
            name_setting=NameSetting("kcb", "", TPL),
            expr="{kex_bc}/(1.0+{keq_bc})",
        ),
        "kcd": ParamLocalSetting(
            name_setting=NameSetting("kcd", "", TPL),
            expr="{keq_cd}*{kex_cd}/(1.0+{keq_cd})",
        ),
        "kdc": ParamLocalSetting(
            name_setting=NameSetting("kdc", "", TPL),
            expr="{kex_cd}/(1.0+{keq_cd})",
        ),
        "pa": ParamLocalSetting(
            name_setting=NameSetting("pa", "", TPL),
            expr=f"{{c_p}} / {p_total}",
        ),
        "pb": ParamLocalSetting(
            name_setting=NameSetting("pb", "", TPL),
            expr=f"{{c_pl1}} / {p_total}",
        ),
        "pc": ParamLocalSetting(
            name_setting=NameSetting("pc", "", TPL),
            expr=f"{{c_pl2}} / {p_total}",
        ),
        "pd": ParamLocalSetting(
            name_setting=NameSetting("pc", "", TPL),
            expr=f"{{c_pl3}} / {p_total}",
        ),
    }


def register() -> None:
    model_factory.register(
        name=NAME,
        setting_maker=make_settings_4st_binding_3_bound_states,
    )
    user_functions = {
        "concentrations": calculate_concentrations,
    }
    user_function_registry.register(name=NAME, user_functions=user_functions)



from __future__ import annotations

import configparser
import sys
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt

import chemex.parameters.name as cpn
import chemex.parameters.spin_system as cns
from chemex.messages import print_making_plots, print_section

if TYPE_CHECKING:
    from argparse import Namespace


def plot_param(args: Namespace) -> None:
    """Plot values of a parameter versus residue number."""
    params = configparser.ConfigParser()

    if len(args.parameters) > 1:
        sys.exit(
            "\nError: Multiple parameter files were given. 'chemex plot_param' "
            "should only be run with a single parameter file.\n",
        )

    params.read(str(args.parameters.pop()))
    param_name = cpn.ParamName.from_section(args.parname)
    curves = {}

    print_making_plots()

    for section in params.sections():
        section_name = cpn.ParamName.from_section(section.strip('"'))
        if param_name.match(section_name):
            print_section(section)
            residues: list[int] = []
            values: list[float] = []
            errors: list[float] = []
            for key, entry in params.items(section):
                residues.append(int(cns.SpinSystem(key).numbers["i"]))
                split = entry.split()
                values.append(float(split[0]))
                try:
                    error = float(split[2].strip("±"))
                except ValueError:
                    error = 0.0
                errors.append(error)
            curves[section] = (residues, values, errors)

    _, axis = plt.subplots(figsize=(12, 5))
    axis.yaxis.grid(True)

    for section, (residues, values, errors) in curves.items():
        axis.errorbar(
            residues,
            values,
            yerr=errors,
            label=section,
            fmt=".",
            barsabove=True,
        )

    plt.legend()
    plt.show()



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from chemex.containers.data import Data


class Printer(Protocol):
    header: str = ""

    def print(self, name: str, data: Data) -> str:
        ...


@dataclass
class ShiftPrinter:
    header = "\n".join(
        [
            "#",
            "# Unit: PPB",
            "#",
            (
                f"# {'NAME':>16s} {'SHIFT (EXP)':>17s}"
                f" {'ERROR (EXP)':>17s} {'SHIFT (CALC)':>17s}\n"
            ),
        ],
    )

    def print(self, name: str, data: Data) -> str:
        return (
            f"{name:>18s} {data.exp[0]: 17.3f} {data.err[0]: 17.3f}"
            f" {data.calc[0]: 17.3f}\n"
        )


@dataclass
class ProfilePrinter:
    first_column_name: str = ""
    first_column_fmt: str = "12.2f"
    header = ""

    def print(self, name: str, data: Data) -> str:
        output = [
            f"[{name}]",
            (
                f"# {self.first_column_name:>12s} {'INTENSITY (EXP)':>17s}"
                f" {'ERROR (EXP)':>17s} {'INTENSITY (CALC)':>17s}"
            ),
        ]

        for metadata, exp, err, calc, mask in zip(
            data.metadata,
            data.exp,
            data.err,
            data.calc,
            data.mask,
            strict=True,
        ):
            start = " " if mask else "#"
            end = "" if mask else " # NOT USED IN THE FIT"
            output.append(
                f"{start} {metadata: {self.first_column_fmt}} {exp: 17.8e}"
                f" {err: 17.8e} {calc: 17.8e}{end}",
            )
        return "\n".join(output) + "\n\n"


@dataclass
class CestPrinter(ProfilePrinter):
    first_column_name: str = "OFFSET (HZ)"
    first_column_fmt: str = "12.2f"


@dataclass
class CpmgPrinter(ProfilePrinter):
    first_column_name: str = "NCYC"
    first_column_fmt: str = "12.0f"


@dataclass
class RelaxationPrinter(ProfilePrinter):
    first_column_name: str = "TIME (S)"
    first_column_fmt: str = "12.2f"


@dataclass
class EXSYPrinter(ProfilePrinter):
    first_column_name: str = "TIME (S)"
    first_column_fmt: str = "12.6f"
    header = ""

    def print(self, name: str, data: Data) -> str:
        output = [
            f"[{name}]",
            f"# {self.first_column_name:>12s} {'STATE1':>12s} {'STATE2':>12s} {'INTENSITY (EXP)':>17s} {'ERROR (EXP)':>17s} {'INTENSITY (CALC)':>17s}",
        ]

        for metadata, exp, err, calc, mask in zip(
            data.metadata,
            data.exp,
            data.err,
            data.calc,
            data.mask,
        ):
            start = " " if mask else "#"
            end = "" if mask else " # NOT USED IN THE FIT"
            time = metadata["times"]
            state1 = metadata["states1"]
            state2 = metadata["states1"]
            output.append(
                f"{start} {time: {self.first_column_fmt}} {state1} {state2} {exp: 17.8e} {err: 17.8e} {calc: 17.8e}{end}",
            )
        return "\n".join(output) + "\n\n"



from __future__ import annotations

from dataclasses import dataclass
from functools import reduce
from typing import TYPE_CHECKING, Literal

import numpy as np

from chemex.configuration.data import RelaxationDataSettings
from chemex.configuration.experiment import CpmgSettings, ExperimentConfig, ToBeFitted
from chemex.containers.dataset import load_relaxation_dataset
from chemex.experiments.factories import Creators, factories
from chemex.filterers import PlanesFilterer
from chemex.nmr.basis import Basis
from chemex.nmr.liouvillian import LiouvillianIS
from chemex.nmr.spectrometer import Spectrometer
from chemex.plotters.cpmg import CpmgPlotter
from chemex.printers.data import CpmgPrinter

if TYPE_CHECKING:
    from chemex.containers.data import Data
    from chemex.parameters.spin_system import SpinSystem
    from chemex.typing import ArrayBool, ArrayFloat, ArrayInt


Delays = tuple[dict[float, float], dict[float, float], list[float]]


EXPERIMENT_NAME = "cpmg_15n_tr_0013"


class Cpmg15NTr0013Settings(CpmgSettings):
    name: Literal["cpmg_15n_tr_0013"]
    time_t2: float
    carrier: float
    pw90: float
    ncyc_max: int
    time_equil: float = 0.0
    taub: float = 2.68e-3
    antitrosy: bool = False
    s3e: bool = True
    observed_state: Literal["a", "b", "c", "d"] = "a"

    @property
    def t_neg(self) -> float:
        return -2.0 * self.pw90 / np.pi

    @property
    def taub_eff(self) -> float:
        return self.taub - 2.0 * self.pw90 - 2.0 * self.pw90 / np.pi

    @property
    def detection(self) -> str:
        if self.antitrosy:
            return f"[2izsz_{self.observed_state}] + [iz_{self.observed_state}]"
        return f"[2izsz_{self.observed_state}] - [iz_{self.observed_state}]"


class Cpmg15NTr0013Config(
    ExperimentConfig[Cpmg15NTr0013Settings, RelaxationDataSettings],
):
    @property
    def to_be_fitted(self) -> ToBeFitted:
        state = self.experiment.observed_state

        to_be_fitted = ToBeFitted(rates=[f"r2_i_{state}"], model_free=[f"tauc_{state}"])

        if self.experiment.antitrosy:
            to_be_fitted.rates.append(f"etaxy_i_{state}")
            to_be_fitted.model_free.append(f"s2_{state}")

        return to_be_fitted


def build_spectrometer(
    config: Cpmg15NTr0013Config,
    spin_system: SpinSystem,
) -> Spectrometer:
    settings = config.experiment
    conditions = config.conditions

    basis = Basis(type="ixyzsz", spin_system="nh")
    liouvillian = LiouvillianIS(spin_system, basis, conditions)
    spectrometer = Spectrometer(liouvillian)

    spectrometer.carrier_i = settings.carrier
    spectrometer.b1_i = 1 / (4.0 * settings.pw90)
    spectrometer.detection = settings.detection

    return spectrometer


@dataclass
class Cpmg15NTr0013Sequence:
    settings: Cpmg15NTr0013Settings

    def _get_delays(self, ncycs: ArrayFloat) -> Delays:
        ncycs_no_ref = ncycs[ncycs > 0]
        tau_cps = {
            ncyc: self.settings.time_t2 / (4.0 * ncyc) - 0.75 * self.settings.pw90
            for ncyc in ncycs_no_ref
        }
        deltas = {
            ncyc: 0.5 * self.settings.pw90 * (self.settings.ncyc_max - ncyc)
            for ncyc in ncycs_no_ref
        }
        deltas[0.0] = 0.5 * self.settings.pw90 * (self.settings.ncyc_max - 1)
        delays = [
            self.settings.t_neg,
            self.settings.taub_eff,
            self.settings.time_equil,
            *deltas.values(),
            *tau_cps.values(),
        ]
        return tau_cps, deltas, delays

    def _get_phases(self, ncyc: float) -> tuple[ArrayInt, ArrayInt]:
        cp_phases1 = np.array(
            [
                [1, 1, 0, 2, 1, 1, 2, 0, 1, 1, 2, 0, 1, 1, 0, 2],
                [0, 2, 3, 3, 2, 0, 3, 3, 2, 0, 3, 3, 0, 2, 3, 3],
            ],
        )
        cp_phases2 = np.array(
            [
                [0, 0, 1, 3, 0, 0, 3, 1, 0, 0, 3, 1, 0, 0, 1, 3],
                [1, 3, 2, 2, 3, 1, 2, 2, 3, 1, 2, 2, 1, 3, 2, 2],
            ],
        )
        indexes = np.arange(int(ncyc))
        phases1 = np.take(cp_phases1, np.flip(indexes), mode="wrap", axis=1)
        phases2 = np.take(cp_phases2, indexes, mode="wrap", axis=1)
        return phases1, phases2

    def _get_start(self, spectrometer: Spectrometer) -> ArrayFloat:
        start = spectrometer.get_start_magnetization(["2izsz"])
        if self.settings.s3e:
            if self.settings.antitrosy:
                start += spectrometer.get_start_magnetization(["iz"])
            else:
                start -= spectrometer.get_start_magnetization(["iz"])
            start *= 0.5
        return start

    def calculate(self, spectrometer: Spectrometer, data: Data) -> ArrayFloat:
        ncycs = data.metadata

        # Calculation of the spectrometers corresponding to all the delays
        tau_cps, deltas, all_delays = self._get_delays(ncycs)
        delays = dict(zip(all_delays, spectrometer.delays(all_delays), strict=True))
        d_neg = delays[self.settings.t_neg]
        d_eq = delays[self.settings.time_equil]
        d_taub = delays[self.settings.taub_eff]
        d_delta = {ncyc: delays[delay] for ncyc, delay in deltas.items()}
        d_cp = {ncyc: delays[delay] for ncyc, delay in tau_cps.items()}

        # Calculation of the spectrometers corresponding to all the pulses
        p90 = spectrometer.p90_i
        p180 = spectrometer.p180_i
        p180_sx = spectrometer.perfect180_s[0]

        # Getting the starting magnetization
        start = self._get_start(spectrometer)

        # Calculating the p-element
        if self.settings.antitrosy:
            p180_is = p90[1] @ p90[0] @ p180_sx @ p90[0] @ p90[1]
        else:
            p180_is = p90[0] @ p90[1] @ p180_sx @ p90[1] @ p90[0]
        palmer = np.mean(
            p90[[0, 2]] @ d_taub @ p180_is @ d_taub @ p180_sx @ p90[[1, 3]],
            axis=0,
        )

        # Calculating the instensities as a function of ncyc
        intst = {
            0.0: spectrometer.detect(
                d_eq
                @ d_delta[0]
                @ p90[3]
                @ p180[[1, 0]]
                @ palmer
                @ p180[[0, 1]]
                @ p90[0]
                @ d_delta[0]
                @ start,
            ),
        }
        for ncyc in set(ncycs) - {0.0}:
            phases1, phases2 = self._get_phases(ncyc)
            echo = d_cp[ncyc] @ p180 @ d_cp[ncyc]
            cpmg1 = d_neg @ reduce(np.matmul, echo[phases1.T]) @ d_neg
            cpmg2 = d_neg @ reduce(np.matmul, echo[phases2.T]) @ d_neg
            intst[ncyc] = spectrometer.detect(
                d_eq
                @ d_delta[ncyc]
                @ p90[3]
                @ cpmg2
                @ palmer
                @ cpmg1
                @ p90[0]
                @ d_delta[ncyc]
                @ start,
            )

        # Return profile
        return np.array([intst[ncyc] for ncyc in ncycs])

    @staticmethod
    def is_reference(metadata: ArrayFloat) -> ArrayBool:
        return metadata == 0


def register() -> None:
    creators = Creators(
        config_creator=Cpmg15NTr0013Config,
        spectrometer_creator=build_spectrometer,
        sequence_creator=Cpmg15NTr0013Sequence,
        dataset_creator=load_relaxation_dataset,
        filterer_creator=PlanesFilterer,
        printer_creator=CpmgPrinter,
        plotter_creator=CpmgPlotter,
    )
    factories.register(type=EXPERIMENT_NAME, creators=creators)



"""This module offers a framework for managing NMR analysis parameters.

It includes classes and methods for indexing, updating, and retrieving parameters,
handling parameter constraints, and integration with the lmfit library for
optimization. Additionally, it supports grid definition for parameter exploration and
ensures physical validity of J couplings.
"""
from __future__ import annotations

import re
import sys
from collections import Counter, defaultdict
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import numpy as np
from lmfit import Parameters as ParametersLF

from chemex.messages import (
    print_error_constraints,
    print_error_grid_settings,
    print_status_changes,
    print_warning_negative_jch,
    print_warning_positive_jnh,
)
from chemex.models.model import model
from chemex.nmr.rates import rate_functions
from chemex.parameters.name import ParamName
from chemex.parameters.userfunctions import user_function_registry

if TYPE_CHECKING:
    from collections.abc import Hashable, Iterable, Sequence

    from chemex.configuration.methods import Method
    from chemex.configuration.parameters import DefaultListType
    from chemex.parameters.setting import Parameters, ParamSetting
    from chemex.typing import ArrayFloat


_PARAM_NAME = r"\[(.+?)\]"
_FLOAT = r"[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?"
_LINEAR = rf"lin[(](?P<start>{_FLOAT}),(?P<end>{_FLOAT}),(?P<num>\d+)[)]$"
_GEOMETRIC = rf"log[(](?P<start>{_FLOAT}),(?P<end>{_FLOAT}),(?P<num>\d+)[)]$"
_LIST = rf"([(](({_FLOAT})(,|[)]$))+)"
_GRID_DEFINTION = (_LINEAR, _GEOMETRIC, _LIST)


class ParameterIndex:
    """Maintains an index of parameter names for efficient searching and retrieval.

    This class creates an index structure that allows for quick lookup of parameter
    IDs based on parameter names, which is particularly useful in large datasets
    with numerous parameters.

    Attributes:
        _index (defaultdict[Hashable, set[str]]): Maps search keys to parameter IDs.
    """

    def __init__(self) -> None:
        """Initializes an empty parameter index."""
        self._index: defaultdict[Hashable, set[str]] = defaultdict(set)

    def add(self, param_name: ParamName) -> None:
        """Adds a parameter name to the index.

        Args:
            param_name (ParamName): The parameter name to be indexed.
        """
        for search_key in param_name.search_keys:
            self._index[search_key].add(param_name.id_)

    def get_matching_ids(self, param_name: ParamName) -> set[str]:
        """Retrieves parameter IDs matching a given parameter name.

        Args:
            param_name (ParamName): Parameter name to match in the index.

        Returns:
            set[str]: Set of matching parameter IDs.
        """
        search_keys = param_name.search_keys
        return set[str].intersection(
            *(self._index.get(search_key, set()) for search_key in search_keys),
        )


def _convert_grid_expression_to_values(grid_expression: str) -> ArrayFloat:
    """Converts grid expression to floating-point values.

    Parses grid expressions used for parameter exploration and converts them into
    an array of floating-point values for further processing.

    Args:
        grid_expression (str): Grid expression to be converted.

    Returns:
        ArrayFloat: Array of floating-point values from the grid expression.
    """
    if match := re.match(_LINEAR, grid_expression):
        return np.linspace(
            float(match.group("start")),
            float(match.group("end")),
            int(match.group("num")),
        )
    if match := re.match(_GEOMETRIC, grid_expression):
        return np.geomspace(
            float(match.group("start")),
            float(match.group("end")),
            int(match.group("num")),
        )
    return np.fromstring(grid_expression.strip("(){}[]"), sep=",")


@dataclass
class ParameterCatalog:
    """Manages a collection of parameter settings for NMR data analysis.

    This class is responsible for storing, updating, and retrieving various
    parameter settings, including their values, expressions, and dependencies.
    It also interfaces with the lmfit library for optimization purposes.

    Attributes:
        _parameters (Parameters): The parameters managed by the catalog.
        _index (ParameterIndex): An index for efficient parameter lookup.
    """

    _parameters: Parameters = field(default_factory=dict)
    _index: ParameterIndex = field(default_factory=ParameterIndex)

    def _add(self, parameter: ParamSetting) -> None:
        """Adds a single parameter to the catalog.

        Args:
            parameter (ParamSetting): The parameter to be added.
        """
        if parameter.id_ not in self._parameters:
            self._parameters[parameter.id_] = parameter
            self._index.add(parameter.param_name)

        if parameter.vary:
            self._parameters[parameter.id_].vary = True
            self._parameters[parameter.id_].expr = ""

    def add_multiple(self, parameters: Parameters) -> None:
        """Adds multiple parameters to the catalog.

        Args:
            parameters (Parameters): Dictionary of parameters to be added.
        """
        for parameter in parameters.values():
            self._add(parameter)

    def get_parameters(self, param_ids: Iterable[str]) -> Parameters:
        """Retrieves a subset of parameters based on their IDs.

        Args:
            param_ids (Iterable[str]): IDs of the parameters to retrieve.

        Returns:
            Parameters: Subset of parameters matching the provided IDs.
        """
        relevant_ids: set[str] = set()

        pool_ids = set(self._parameters) & set(param_ids)

        while pool_ids:
            for param_id in pool_ids.copy():
                relevant_ids.add(param_id)
                pool_ids.discard(param_id)
                pool_ids.update(self._parameters[param_id].dependencies)
            pool_ids -= relevant_ids

        return {
            param_id: parameter
            for param_id, parameter in self._parameters.items()
            if param_id in relevant_ids
        }

    def build_lmfit_params(
        self,
        param_ids: Iterable[str] | None = None,
    ) -> ParametersLF:
        """Constructs lmfit Parameters from the catalog's parameters.

        This method converts the internal parameter representation to the format
        required by the lmfit library for optimization.

        Args:
            param_ids (Iterable[str] | None): Specific parameter IDs to include.
                                              If None, includes all parameters.

        Returns:
            ParametersLF: lmfit-compatible parameter collection.
        """
        if param_ids is None:
            param_ids = set(self._parameters)

        parameters = self.get_parameters(param_ids)

        parameter_args = (parameter.args for parameter in parameters.values())

        usersyms = rate_functions | user_function_registry.get(model.name)
        lmfit_params = ParametersLF(usersyms=usersyms)
        lmfit_params.add_many(*parameter_args)
        lmfit_params.update_constraints()

        for id_, lf_param in lmfit_params.items():
            lf_param.stderr = parameters[id_].stderr

        return lmfit_params

    def update_from_lmfit_params(self, parameters: ParametersLF) -> None:
        """Updates catalog parameters from lmfit Parameters.

        Args:
            parameters (ParametersLF): lmfit Parameters to update from.
        """
        for param_id, parameter in parameters.items():
            self._parameters[param_id].value = parameter.value
            self._parameters[param_id].stderr = parameter.stderr

    def get_matching_ids(self, param_name: ParamName) -> set[str]:
        """Finds parameter IDs matching a specified parameter name.

        Args:
            param_name (ParamName): Name of the parameter to match.

        Returns:
            set[str]: Set of IDs matching the parameter name.
        """
        return self._index.get_matching_ids(param_name)

    def get_value(self, id_: str) -> float | None:
        """Retrieves the value of a parameter by its ID.

        Args:
            id_ (str): The ID of the parameter.

        Returns:
            float | None: The value of the parameter, or None if not found.
        """
        return self._parameters[id_].value

    def set_values(self, par_values: dict[str, float]) -> None:
        """Sets the values of specified parameters.

        Args:
            par_values (dict[str, float]): Dictionary of parameter IDs and values.
        """
        for id_, value in par_values.items():
            if id_ in self._parameters:
                self._parameters[id_].value = value

    def set_defaults(self, defaults: DefaultListType) -> None:
        """Sets default settings for a group of parameters.

        Args:
            defaults (DefaultListType): Defaults to apply.
        """
        id_pool = set(self._parameters)
        for name_to_set, setting in reversed(defaults):
            matching_ids = self.get_matching_ids(name_to_set) & id_pool
            id_pool -= matching_ids
            for matching_id in matching_ids:
                self._parameters[matching_id].set(setting)

    def _count_per_section(self, param_ids: set[str]) -> Counter[str]:
        """Counts parameters per section for a given set of IDs.

        Args:
            param_ids (set[str]): Set of parameter IDs to count.

        Returns:
            Counter[str]: Counts of parameters per section.
        """
        return Counter(
            self._parameters[param_id].param_name.section for param_id in param_ids
        )

    def set_vary(self, section_names: Sequence[str], vary: bool) -> Counter[str]:
        """Sets the variability of parameters by section name.

        Args:
            section_names (Sequence[str]): Names of the sections to update.
            vary (bool): Whether to set parameters as variable.

        Returns:
            Counter[str]: Counts of parameters updated per section.
        """
        parameters = self._parameters

        ids_modified: set[str] = set()
        ids_pool = {
            param_id
            for param_id, setting in parameters.items()
            if setting.vary != vary or setting.expr
        }

        for section_name in reversed(section_names):
            param_name = ParamName.from_section(section_name)
            matching_ids = self.get_matching_ids(param_name) & ids_pool
            for matching_id in matching_ids:
                parameters[matching_id].vary = vary
                if vary:
                    parameters[matching_id].expr = ""
            ids_modified.update(matching_ids)

        return self._count_per_section(ids_modified)

    def fix_all(self) -> None:
        """Fixes all parameters, preventing them from varying during fitting."""
        for parameter in self._parameters.values():
            parameter.vary = False
            parameter.expr = ""

    def _get_ids_left(self, expression: str) -> set[str]:
        """Extracts parameter IDs from the left side of an expression.

        Args:
            expression (str): The expression to parse.

        Returns:
            set[str]: Set of parameter IDs extracted.
        """
        return self.get_matching_ids(ParamName.from_section(expression))

    def _get_ids_right(self, expression: str) -> dict[str, set[str]]:
        """Extracts parameter IDs from the right side of an expression.

        Args:
            expression (str): The expression to parse.

        Returns:
            dict[str, set[str]]: Mapping of section names to parameter IDs.
        """
        ids_right: dict[str, set[str]] = {}
        for match in re.finditer(_PARAM_NAME, expression):
            param_name = ParamName.from_section(match.group(1))
            ids_right[match.group(0)] = self.get_matching_ids(param_name)
        return ids_right

    def _set_expression(self, expression: str, ids_pool: set[str]) -> set[str]:
        """Sets expressions for parameters based on an input expression.

        Args:
            expression (str): The expression to apply.
            ids_pool (set[str]): Set of parameter IDs to consider.

        Returns:
            set[str]: Set of parameter IDs updated.
        """
        left, right, *something_else = expression.split("=")

        if something_else:
            print_error_constraints(expression)
            sys.exit()

        ids_left = self._get_ids_left(left) & ids_pool
        ids_right_dict = self._get_ids_right(right)

        for id_left in ids_left:
            new_expression = right.strip()
            for section_name, ids_right in ids_right_dict.items():
                name = self._parameters[id_left].param_name
                id_replace = name.get_closest_id(ids_right)
                new_expression = new_expression.replace(section_name, id_replace)
            self._parameters[id_left].expr = new_expression

        return ids_left

    def set_expressions(self, expression_list: Sequence[str]) -> Counter[str]:
        """Applies a list of expressions to parameters.

        Args:
            expression_list (Sequence[str]): Expressions to apply.

        Returns:
            Counter[str]: Counts of parameters updated per section.
        """
        ids_modified: set[str] = set()
        ids_pool = set(self._parameters)

        for expression in reversed(expression_list):
            ids_changed = self._set_expression(expression, ids_pool)
            ids_pool -= ids_changed
            ids_modified.update(ids_changed)

        return self._count_per_section(ids_modified)

    def parse_grid(self, grid_entries: list[str]) -> dict[str, ArrayFloat]:
        """Parses grid definitions and sets up parameters accordingly.

        Args:
            grid_entries (list[str]): List of grid definitions.

        Returns:
            dict[str, ArrayFloat]: Mapping of parameter IDs to grid values.
        """
        ids_pool = set(self._parameters)

        grid_values: dict[str, ArrayFloat] = {}

        for entry in reversed(grid_entries):
            name, expression, *something_else = entry.replace(" ", "").split("=")

            valid_definition = any(
                re.match(regex, expression) for regex in _GRID_DEFINTION
            )

            if something_else or not valid_definition:
                print_error_grid_settings(entry)
                sys.exit()

            ids_left = self.get_matching_ids(ParamName.from_section(name))
            values = _convert_grid_expression_to_values(expression)

            grid_values |= {param_id: values for param_id in ids_left & ids_pool}

            self.set_vary([name], False)

        return grid_values

    def check_params(self):
        """Checks and warns about the physical validity of J couplings."""
        positive_jnh = False
        negative_jch = False
        for setting in self._parameters.values():
            param_name = setting.param_name
            if not param_name.name.startswith("J_") or setting.value is None:
                continue
            atoms = {atom.name for atom in param_name.spin_system.atoms.values()}
            if atoms == {"N", "H"} and setting.value > 0:
                positive_jnh = True
            if atoms == {"C", "H"} and setting.value < 0:
                negative_jch = True
        if positive_jnh:
            print_warning_positive_jnh()
        if negative_jch:
            print_warning_negative_jch()

    def sort(self) -> None:
        """Sorts the parameters in the catalog by their names."""
        sorted_items = sorted(
            self._parameters.items(),
            key=lambda item: item[1].param_name,
        )
        self._parameters = dict(sorted_items)


@dataclass
class ParamManager:
    """Manages a collection of ParameterCatalogs for different fitting models.

    This class acts as an interface to multiple ParameterCatalogs, allowing
    for easy management of parameters across different fitting models used
    in NMR data analysis.

    Attributes:
        _database (ParameterCatalog): The primary parameter catalog.
        _database_mf (ParameterCatalog): The model-free parameter catalog.
    """

    _database: ParameterCatalog
    _database_mf: ParameterCatalog

    @property
    def database(self) -> ParameterCatalog:
        """Returns the appropriate parameter catalog based on the model.

        Returns:
            ParameterCatalog: The active parameter catalog.
        """
        return self._database_mf if model.model_free else self._database

    def add_multiple(self, parameters: Parameters) -> None:
        """Adds multiple parameters to the primary catalog.

        Args:
            parameters (Parameters): Parameters to be added.
        """
        self._database.add_multiple(parameters)

    def add_multiple_mf(self, parameters: Parameters) -> None:
        """Adds multiple parameters to the model-free catalog.

        Args:
            parameters (Parameters): Parameters to be added.
        """
        self._database_mf.add_multiple(parameters)

    def get_parameters(self, param_ids: Iterable[str]) -> Parameters:
        """Retrieves parameters from the active catalog by IDs.

        Args:
            param_ids (Iterable[str]): IDs of the parameters to retrieve.

        Returns:
            Parameters: Retrieved parameters.
        """
        return self.database.get_parameters(param_ids)

    def get_value(self, param_id: str) -> float | None:
        """Gets the value of a parameter by its ID from the active catalog.

        Args:
            param_id (str): ID of the parameter.

        Returns:
            float | None: The value of the parameter or None if not found.
        """
        return self.database.get_value(param_id)

    def build_lmfit_params(self, param_ids: Iterable[str]) -> ParametersLF:
        """Builds lmfit Parameters from the active catalog.

        Args:
            param_ids (Iterable[str]): IDs of parameters to include.

        Returns:
            ParametersLF: lmfit-compatible parameters.
        """
        return self.database.build_lmfit_params(param_ids)

    def sort(self) -> None:
        """Sorts parameters in the active catalog."""
        self.database.sort()

    def update_from_parameters(self, parameters: ParametersLF) -> None:
        """Updates the active catalog from lmfit Parameters.

        Args:
            parameters (ParametersLF): lmfit Parameters to update from.
        """
        self.database.update_from_lmfit_params(parameters)

    def set_values(self, par_values: dict[str, float]) -> None:
        """Sets the values of specific parameters in the active catalog.

        Args:
            par_values (dict[str, float]): Parameter values to set.
        """
        return self.database.set_values(par_values)

    def set_defaults(self, defaults: DefaultListType) -> None:
        """Sets defaults for parameters in both catalogs.

        Args:
            defaults (DefaultListType): Default settings to apply.
        """
        self._database_mf.set_defaults(defaults)

        if model.model_free:
            return

        params_mf = self._database_mf.build_lmfit_params()
        self.database.set_values(params_mf.valuesdict())

        self.database.set_defaults(defaults)

        self.database.check_params()

    def set_vary(self, section_names: Sequence[str], vary: bool) -> Counter[str]:
        """Sets variability of parameters in the active catalog by section name.

        Args:
            section_names (Sequence[str]): Section names to update.
            vary (bool): Whether parameters should vary.

        Returns:
            Counter[str]: Counts of updated parameters per section.
        """
        return self.database.set_vary(section_names, vary)

    def fix_all(self) -> None:
        """Fixes all parameters in the active catalog, preventing them from varying."""
        self.database.fix_all()

    def set_expressions(self, expression_list: Sequence[str]) -> Counter[str]:
        """Applies expressions to parameters in the active catalog.

        Args:
            expression_list (Sequence[str]): Expressions to apply.

        Returns:
            Counter[str]: Counts of updated parameters per section.
        """
        return self.database.set_expressions(expression_list)

    def parse_grid(self, grid_entries: list[str]) -> dict[str, ArrayFloat]:
        """Parses grid definitions and sets up parameters in the active catalog.

        Args:
            grid_entries (list[str]): Grid definitions to parse.

        Returns:
            dict[str, ArrayFloat]: Parameters mapped to grid values.
        """
        return self.database.parse_grid(grid_entries)


_parameter_catalog = ParameterCatalog()
_parameter_catalog_mf = ParameterCatalog()
_manager = ParamManager(_parameter_catalog, _parameter_catalog_mf)

set_param_vary = _manager.set_vary
set_param_expressions = _manager.set_expressions
add_parameters = _manager.add_multiple
add_parameters_mf = _manager.add_multiple_mf
get_parameters = _manager.get_parameters
build_lmfit_params = _manager.build_lmfit_params
update_from_parameters = _manager.update_from_parameters
parse_grid = _manager.parse_grid
set_param_values = _manager.set_values
set_param_defaults = _manager.set_defaults
sort_parameters = _manager.sort
fix_all_parameters = _manager.fix_all


def set_parameter_status(method: Method):
    """Sets parameter status based on a given method configuration.

    This function configures parameters to vary, be fixed, or follow certain
    expressions based on the provided method configuration.

    Args:
        method (Method): Method configuration to apply.
    """
    matches_con = set_param_expressions(method.constraints)
    matches_fix = set_param_vary(method.fix, vary=False)
    matches_fit = set_param_vary(method.fit, vary=True)

    print_status_changes(matches_fit, matches_fix, matches_con)



from enum import Enum, auto


class Nucleus(Enum):
    """Enumeration of different types of atomic nuclei.

    This Enum class represents different types of atomic nuclei commonly
    used in NMR spectroscopy.

    Attributes:
        H1: Represents Hydrogen-1 nucleus.
        N15: Represents Nitrogen-15 nucleus.
        C13: Represents Carbon-13 nucleus.
        X: Placeholder for unknown or unspecified nucleus type.
    """

    H1 = auto()
    N15 = auto()
    C13 = auto()
    X = auto()


# Dictionary mapping atom letter abbreviations to corresponding Nucleus enum values.
STR_TO_NUCLEUS: dict[str, Nucleus] = {
    "H": Nucleus.H1,
    "Q": Nucleus.H1,  # Q, M are aliases for Hydrogen-1
    "M": Nucleus.H1,
    "N": Nucleus.N15,
    "C": Nucleus.C13,
    "X": Nucleus.X,
}


def str2nucleus(atom_letter: str) -> Nucleus:
    """Converts a single-letter atom name to the corresponding Nucleus enum.

    This function maps a single-letter abbreviation for an atom (e.g., 'H', 'C')
    to the corresponding Nucleus enumeration member. If the atom letter is not
    recognized, it defaults to 'X', representing an unknown or unspecified nucleus.

    Args:
        atom_letter (str): A single-letter abbreviation of an atom.

    Returns:
        Nucleus: The corresponding Nucleus enumeration member.

    Examples:
        >>> str2nucleus('H')
        <Nucleus.H1: 1>
        >>> str2nucleus('C')
        <Nucleus.C13: 3>
        >>> str2nucleus('Z')  # Not recognized
        <Nucleus.X: 4>
    """
    return STR_TO_NUCLEUS.get(atom_letter, Nucleus.X)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from matplotlib.figure import Figure

from chemex.containers.data import Data
from chemex.messages import print_plot_filename

if TYPE_CHECKING:
    from pathlib import Path

    from chemex.containers.profile import Profile

_RED700 = "#D32F2F"


def plot_shift(name_pdf: Path, data: Data):
    fig = Figure()
    ax = fig.subplots(1, 1)
    fig.align_labels()
    ax.errorbar(data.calc, data.exp, yerr=data.err, fmt=".", color=_RED700)
    val_min = min(ax.get_xlim()[0], ax.get_ylim()[0])
    val_max = max(ax.get_xlim()[1], ax.get_ylim()[1])
    ax.set_aspect("equal", "box")
    ax.plot([val_min, val_max], [val_min, val_max], color="k", linewidth=0.5, zorder=1)
    ax.set_xlabel(r"$Δδ_\mathregular{fit}$ (ppb)")
    ax.set_ylabel(r"$Δδ_\mathregular{exp}$ (ppb)")
    fig.savefig(str(name_pdf))


def create_plot_data(profiles: list[Profile]) -> Data:
    exp = np.array([profile.data.exp for profile in profiles]).flatten()
    err = np.array([profile.data.err for profile in profiles]).flatten()
    calc = np.array([profile.data.calc for profile in profiles]).flatten()

    data_plot = Data(exp=exp, err=err)
    data_plot.calc = calc

    return data_plot


class ShiftPlotter:
    def __init__(self, filename: Path, **_extra: Any) -> None:
        self.filename = filename
        self.plot_simulation = self.plot

    def plot(self, path: Path, profiles: list[Profile]) -> None:
        basename = path / self.filename.name
        name_pdf = basename.with_suffix(".pdf")

        print_plot_filename(name_pdf, extra=False)

        data_plot = create_plot_data(profiles)
        plot_shift(name_pdf, data_plot)



from __future__ import annotations

from itertools import combinations
from typing import TYPE_CHECKING

from chemex.models.constraints import pop_4st
from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting
from chemex.parameters.userfunctions import user_function_registry

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions

NAME = "4st_hd"

TPL = ("temperature", "p_total", "l_total")


def create_pop_4st_eyring_settings() -> dict[str, ParamLocalSetting]:
    arguments = ", ".join(
        f"{{k{i}{j}}}, {{k{j}{i}}}" for i, j in combinations("abcd", 2)
    )
    return {
        f"p{state}": ParamLocalSetting(
            name_setting=NameSetting(f"p{state}", "", TPL),
            expr=(f"pop_4st({arguments})['p{state}']"),
        )
        for state in "abcd"
    }


def make_settings_4st_hd(conditions: Conditions) -> dict[str, ParamLocalSetting]:
    d2o: float = conditions.d2o if conditions.d2o is not None else 0.1
    return {
        "d2o": ParamLocalSetting(
            name_setting=NameSetting("d2o", "", ("d2o",)),
            value=d2o,
            min=0.0,
            max=1.0,
        ),
        "pop_b": ParamLocalSetting(
            name_setting=NameSetting("pop_b", "", TPL),
            value=0.02,
            min=0.0,
            max=1.0,
            vary=True,
        ),
        "kex_ab": ParamLocalSetting(
            name_setting=NameSetting("kex_ab", "", TPL),
            min=0.0,
            value=0.0,
            vary=True,
        ),
        "kdh_a": ParamLocalSetting(
            name_setting=NameSetting("kdh_a", "g", ("temperature",)),
            value=1.0,
            min=0.0,
            vary=True,
        ),
        "kdh_b": ParamLocalSetting(
            name_setting=NameSetting("kdh_b", "g", ("temperature",)),
            value=1.0,
            min=0.0,
            vary=True,
        ),
        "phi_a": ParamLocalSetting(
            name_setting=NameSetting("phi_a", "g", ("temperature",)),
            value=1.1,
            min=0.75,
            max=1.50,
        ),
        "phi_b": ParamLocalSetting(
            name_setting=NameSetting("phi_b", "g", ("temperature",)),
            value=0.02,
            min=0.75,
            max=1.50,
            expr="{phi_a}",
        ),
        "kab": ParamLocalSetting(
            name_setting=NameSetting("kab", "", TPL),
            min=0.0,
            expr="{pop_b} * {kex_ab}",
        ),
        "kba": ParamLocalSetting(
            name_setting=NameSetting("kba", "", TPL),
            min=0.0,
            expr="(1.0 - {pop_b}) * {kex_ab}",
        ),
        "kcd": ParamLocalSetting(
            name_setting=NameSetting("kcd", "", TPL),
            min=0.0,
            expr="{pop_b} * {kex_ab}",
        ),
        "kdc": ParamLocalSetting(
            name_setting=NameSetting("kdc", "", TPL),
            min=0.0,
            expr="(1.0 - {pop_b}) * {kex_ab}",
        ),
        "kac": ParamLocalSetting(
            name_setting=NameSetting("kac", "g", ("temperature", "d2o")),
            min=0.0,
            expr="{d2o} * {kdh_a} * {phi_a}",
        ),
        "kca": ParamLocalSetting(
            name_setting=NameSetting("kca", "g", ("temperature", "d2o")),
            min=0.0,
            expr="(1.0 - {d2o}) * {kdh_a}",
        ),
        "kbd": ParamLocalSetting(
            name_setting=NameSetting("kbd", "g", ("temperature", "d2o")),
            min=0.0,
            expr="{d2o} * {kdh_b} * {phi_b}",
        ),
        "kdb": ParamLocalSetting(
            name_setting=NameSetting("kdb", "g", ("temperature", "d2o")),
            min=0.0,
            expr="(1.0 - {d2o}) * {kdh_b}",
        ),
        **create_pop_4st_eyring_settings(),
    }


def register() -> None:
    model_factory.register(name=NAME, setting_maker=make_settings_4st_hd)
    user_function_registry.register(name=NAME, user_functions={"pop_4st": pop_4st})



from __future__ import annotations

from contextlib import ExitStack
from typing import TYPE_CHECKING, Any, Generic, Protocol, TypeVar

import numpy as np
from matplotlib.backends.backend_pdf import PdfPages

from chemex.containers.data import Data
from chemex.messages import print_plot_filename
from chemex.plotters.plot import plot_profile
from chemex.printers.plot import PlotPrinter, data_plot_printers

if TYPE_CHECKING:
    from pathlib import Path

    from chemex.containers.profile import Profile
    from chemex.typing import ArrayFloat


LARGE_ERROR = 1e15


rng = np.random.default_rng()


class CpmgExperimentSettings(Protocol):
    time_t2: float
    even_ncycs: bool = False


class CpmgExperimentConfig(Protocol):
    experiment: CpmgExperimentSettings


T = TypeVar("T", bound=CpmgExperimentConfig)


def plot_cpmg(file_pdf: PdfPages, name: str, data_exp: Data, data_calc: Data):
    fig = plot_profile(name, data_exp, data_calc)
    ax2 = fig.axes[1]
    ax2.set_xlabel(r"$\nu_\mathregular{CPMG}$ (Hz)")
    ax2.set_ylabel(r"$R_{2,\mathregular{eff}}$ (s$^{-1}$)")
    ax2.set_xlim(0.0, max(data_calc.metadata) + min(data_calc.metadata))
    file_pdf.savefig(fig)


def intensities_to_rates(
    intensities: ArrayFloat,
    intensities0: ArrayFloat,
    time_t2: float,
) -> ArrayFloat:
    normalized_intensities = intensities / np.mean(intensities0, axis=-1, keepdims=True)

    # If the normalized intensity is negative, no rate can be estimated.
    # The rate is then set to infinity by default.
    rates = np.full_like(intensities, np.inf)
    neg = normalized_intensities <= 0
    rates[~neg] = -np.log(normalized_intensities[~neg]) / time_t2

    return rates


def calculate_exp_rates(data: Data, time_t2: float) -> ArrayFloat:
    intensities = data.exp[~data.refs]
    intensities0 = data.exp[data.refs]
    return intensities_to_rates(intensities, intensities0, time_t2)


def calculate_calc_rates(data: Data, time_t2: float) -> ArrayFloat:
    intensities = data.calc[~data.refs]
    intensities0 = data.exp[data.refs]
    return intensities_to_rates(intensities, intensities0, time_t2)


def calculate_errorbars(data: Data, rates: ArrayFloat, time_t2: float) -> ArrayFloat:
    randn = rng.standard_normal(size=(10000, 1))
    randn0 = rng.standard_normal(size=(10000, 1))

    intensities = data.exp[~data.refs]
    intensities0 = data.exp[data.refs]

    intensities_err = data.err[~data.refs]
    intensities0_err = data.err[data.refs]

    intensities_ensemble = intensities + intensities_err * randn
    intensities0_ensemble = intensities0 + intensities0_err * randn0

    rates_ensemble: ArrayFloat = intensities_to_rates(
        intensities_ensemble,
        intensities0_ensemble,
        time_t2,
    )

    # Set infinity rates to high values as `np.percentile` does not take infinity
    rates_ensemble[rates_ensemble == np.inf] = 1e16

    errors = np.percentile(rates_ensemble - rates, [15.9, 84.1], axis=0).transpose()

    # Set back large errors to infinity
    errors[errors > LARGE_ERROR] = np.inf

    return np.abs(errors)


def ncycs_to_nu_cpmgs(ncycs: ArrayFloat, time_t2: float) -> ArrayFloat:
    modified_ncycs = ncycs.copy()
    modified_ncycs[modified_ncycs == -1] = 0.5
    return modified_ncycs[modified_ncycs != 0] / time_t2


def create_plot_data_exp(profile: Profile, config: CpmgExperimentConfig) -> Data:
    time_t2 = config.experiment.time_t2
    data = profile.data
    refs = data.refs

    nu_cpmgs = ncycs_to_nu_cpmgs(data.metadata, time_t2)
    rates = calculate_exp_rates(data, time_t2)
    rates_calc = calculate_calc_rates(data, time_t2)
    errorbars = calculate_errorbars(data, rates, time_t2)

    data_exp = Data(exp=rates, err=errorbars, metadata=nu_cpmgs)
    data_exp.calc = rates_calc
    data_exp.mask = data.mask[~refs]
    data_exp.sort()

    return data_exp


def create_plot_data_calc(profile: Profile, config: CpmgExperimentConfig) -> Data:
    spectrometer = profile.spectrometer
    time_t2 = config.experiment.time_t2
    data = profile.data
    refs = data.refs

    step = 2 if config.experiment.even_ncycs else 1
    ncycs = np.arange(2, max(data.metadata) + 1, step, dtype=np.float_)
    ncycs = np.asarray(sorted(set(ncycs) | set(data.metadata[~refs])))

    filler = np.zeros_like(ncycs)
    data_for_calculation = Data(exp=filler, err=filler, metadata=ncycs)

    intensities = profile.data.scale * profile.pulse_sequence.calculate(
        spectrometer,
        data_for_calculation,
    )

    nu_cpmgs = ncycs_to_nu_cpmgs(ncycs, time_t2)
    data_fit = Data(exp=filler, err=filler, metadata=nu_cpmgs)
    data_fit.calc = intensities_to_rates(intensities, data.exp[refs], time_t2)

    return data_fit


class CpmgPlotter(Generic[T]):
    def __init__(self, filename: Path, config: T, **_extra: Any) -> None:
        self.filename = filename
        self.config = config
        self.printer: PlotPrinter = data_plot_printers["cpmg"]

    def plot(self, path: Path, profiles: list[Profile]) -> None:
        basename = path / self.filename.name
        name_pdf = basename.with_suffix(".pdf")
        name_exp = basename.with_suffix(".exp")
        name_fit = basename.with_suffix(".fit")

        print_plot_filename(name_pdf)

        with ExitStack() as stack:
            file_pdf = stack.enter_context(PdfPages(str(name_pdf)))
            file_calc = stack.enter_context(name_fit.open("w"))
            file_exp = stack.enter_context(name_exp.open("w"))
            for profile in sorted(profiles):
                data_exp = create_plot_data_exp(profile, self.config)
                data_calc = create_plot_data_calc(profile, self.config)
                plot_cpmg(file_pdf, str(profile.spin_system), data_exp, data_calc)
                file_exp.write(
                    self.printer.print_exp(str(profile.spin_system), data_exp)
                )
                file_calc.write(
                    self.printer.print_calc(str(profile.spin_system), data_calc)
                )

    def plot_simulation(self, path: Path, profiles: list[Profile]) -> None:
        basename = path / self.filename.name
        name_pdf = basename.with_suffix(".pdf")
        name_sim = basename.with_suffix(".sim")

        print_plot_filename(name_pdf, extra=False)

        with ExitStack() as stack:
            file_pdf = stack.enter_context(PdfPages(str(name_pdf)))
            file_sim = stack.enter_context(name_sim.open("w"))
            for profile in sorted(profiles):
                data_exp = create_plot_data_exp(profile, self.config)
                data_calc = create_plot_data_calc(profile, self.config)
                plot_cpmg(file_pdf, str(profile.spin_system), data_exp, data_calc)
                file_sim.write(
                    self.printer.print_calc(str(profile.spin_system), data_calc)
                )



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from chemex.containers.data import Data
from chemex.parameters.spin_system import SpinSystem
from chemex.toml import normalize_path

if TYPE_CHECKING:
    from pathlib import Path

    from chemex.configuration.data import RelaxationDataSettings, ShiftDataSettings
    from chemex.configuration.experiment import ExperimentConfig, ExperimentNameSettings

    # Type aliases
    Dataset = list[tuple[SpinSystem, Data]]
    RelaxationConfig = ExperimentConfig[ExperimentNameSettings, RelaxationDataSettings]
    ShiftConfig = ExperimentConfig[ExperimentNameSettings, ShiftDataSettings]


def load_relaxation_dataset(base_path: Path, settings: RelaxationConfig) -> Dataset:
    data_path = normalize_path(base_path, settings.data.path)
    dtype = [("metadata", "f8"), ("exp", "f8"), ("err", "f8")]

    dataset: Dataset = []
    for spin_system, filepaths in settings.data.profiles.items():
        for filepath in filepaths:
            raw_data = np.loadtxt(data_path / filepath, dtype=dtype, usecols=[0, 1, 2])
            dataset.append(
                (
                    spin_system,
                    Data(
                        exp=raw_data["exp"],
                        err=raw_data["err"],
                        metadata=raw_data["metadata"],
                    ),
                ),
            )

    return dataset


def load_exsy_dataset(base_path: Path, settings: RelaxationConfig) -> Dataset:
    data_path = normalize_path(base_path, settings.data.path)
    dtype = [
        ("times", "f8"),
        ("states1", "U1"),
        ("states2", "U1"),
        ("exp", "f8"),
        ("err", "f8"),
    ]

    dataset: Dataset = []
    for spin_system, filepaths in settings.data.profiles.items():
        for filepath in filepaths:
            raw_data = np.loadtxt(data_path / filepath, dtype=dtype)
            dataset.append(
                (
                    spin_system,
                    Data(
                        exp=raw_data["exp"],
                        err=raw_data["err"],
                        metadata=raw_data[["times", "states1", "states2"]],
                    ),
                ),
            )

    return dataset


def load_shift_dataset(base_path: Path, settings: ShiftConfig) -> Dataset:
    data_path = normalize_path(base_path, settings.data.path)

    shifts = np.loadtxt(
        data_path,
        dtype=[("spin_system", "U15"), ("exp", "f8"), ("err", "f8")],
    )

    return [
        (SpinSystem(spin_system), Data(exp=np.array([exp]), err=np.array([err])))
        for spin_system, exp, err in shifts
    ]



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal

import numpy as np

from chemex.configuration.data import CestDataSettings
from chemex.configuration.experiment import CestSettings, ExperimentConfig, ToBeFitted
from chemex.containers.dataset import load_relaxation_dataset
from chemex.experiments.factories import Creators, factories
from chemex.filterers import CestFilterer
from chemex.nmr.basis import Basis
from chemex.nmr.constants import get_multiplet
from chemex.nmr.liouvillian import LiouvillianIS
from chemex.nmr.spectrometer import Spectrometer
from chemex.plotters.cest import CestPlotter
from chemex.printers.data import CestPrinter

if TYPE_CHECKING:
    from chemex.containers.data import Data
    from chemex.parameters.spin_system import SpinSystem
    from chemex.typing import ArrayBool, ArrayFloat


EXPERIMENT_NAME = "cest_13c"
OFFSET_REF = 1e4


class Cest13CSettings(CestSettings):
    name: Literal["cest_13c"]
    time_t1: float
    carrier: float
    b1_frq: float
    b1_inh_scale: float = 0.1
    b1_inh_res: int = 11
    observed_state: Literal["a", "b", "c", "d"] = "a"

    @property
    def detection(self) -> str:
        return f"[iz_{self.observed_state}]"


class Cest13CConfig(ExperimentConfig[Cest13CSettings, CestDataSettings]):
    @property
    def to_be_fitted(self) -> ToBeFitted:
        state = self.experiment.observed_state
        return ToBeFitted(
            rates=["r2_i", f"r1_i_{state}"],
            model_free=[f"tauc_{state}", f"s2_{state}"],
        )


def build_spectrometer(config: Cest13CConfig, spin_system: SpinSystem) -> Spectrometer:
    settings = config.experiment
    conditions = config.conditions

    basis = Basis(type="ixyz", spin_system="ch")
    liouvillian = LiouvillianIS(spin_system, basis, conditions)
    spectrometer = Spectrometer(liouvillian)

    spectrometer.carrier_i = settings.carrier
    spectrometer.b1_i = settings.b1_frq
    spectrometer.b1_i_inh_scale = settings.b1_inh_scale
    spectrometer.b1_i_inh_res = settings.b1_inh_res

    spectrometer.detection = settings.detection

    if "13c" in conditions.label:
        symbol = spin_system.symbols["i"]
        atom = spin_system.atoms["i"]
        liouvillian.jeff_i = get_multiplet(symbol, atom.name)

    return spectrometer


@dataclass
class Cest13CSequence:
    settings: Cest13CSettings

    @staticmethod
    def is_reference(metadata: ArrayFloat) -> ArrayBool:
        return np.abs(metadata) > OFFSET_REF

    def calculate(self, spectrometer: Spectrometer, data: Data) -> ArrayFloat:
        offsets = data.metadata

        start = spectrometer.get_equilibrium()

        intensities: dict[float, ArrayFloat] = {}

        for offset in set(offsets):
            intensities[offset] = start

            if self.is_reference(offset):
                continue

            spectrometer.offset_i = offset

            intensities[offset] = (
                spectrometer.pulse_i(self.settings.time_t1, 0.0) @ intensities[offset]
            )

        return np.array(
            [spectrometer.detect(intensities[offset]) for offset in offsets],
        )


def register() -> None:
    creators = Creators(
        config_creator=Cest13CConfig,
        spectrometer_creator=build_spectrometer,
        sequence_creator=Cest13CSequence,
        dataset_creator=load_relaxation_dataset,
        filterer_creator=CestFilterer,
        printer_creator=CestPrinter,
        plotter_creator=CestPlotter,
    )
    factories.register(type=EXPERIMENT_NAME, creators=creators)



import numpy as np
from scipy.interpolate import CubicSpline

from chemex.containers.profile import Profile
from chemex.plotters.cest import create_plot_data_exp


class Curve:
    def __init__(self, profile: Profile, sw: float | None = None) -> None:
        data = create_plot_data_exp(profile)
        settings = getattr(profile.pulse_sequence, "settings", None)

        delta_ppm = 0.0
        if settings is not None:
            spectrometer = profile.spectrometer
            cos_n = getattr(settings, "cos_n", None)
            sw = getattr(settings, "sw", None)
            if cos_n is not None and cos_n % 2 == 0 and sw is not None:
                shifts_ppm = spectrometer.offsets_to_ppms(np.array([sw / 2.0, 0.0]))
                delta_ppm = shifts_ppm[0] - shifts_ppm[1]

        ppms = data.metadata - delta_ppm
        xmin, xmax = min(ppms), max(ppms)
        self.xrange = xmax - xmin
        self.xcentre = 0.5 * (xmin + xmax)
        self.x = ppms
        self.y = data.exp
        _, indices = np.unique(ppms, return_index=True)
        x = ppms[indices]
        y = data.exp[indices]
        if sw is not None:
            y[0] = y[-1] = 0.5 * (y[0] + y[-1])
            self.spline = CubicSpline(x, y, bc_type="periodic", extrapolate="periodic")
        else:
            self.spline = CubicSpline(x, y)

    def get_xrange(self, sw: float | None = None):
        if sw is None:
            sw = 1.0
        return self.xcentre + sw * self.xrange * np.array([-0.5, 0.5])



from __future__ import annotations

from typing import TYPE_CHECKING

from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions


NAME = "2st"

TPL = ("temperature", "p_total", "l_total")


def make_settings_2st(_conditions: Conditions) -> dict[str, ParamLocalSetting]:
    return {
        "kex_ab": ParamLocalSetting(
            name_setting=NameSetting("kex_ab", "", TPL),
            value=200.0,
            min=0.0,
            vary=True,
        ),
        "pb": ParamLocalSetting(
            name_setting=NameSetting("pb", "", TPL),
            value=0.05,
            min=0.0,
            max=1.0,
            vary=True,
        ),
        "pa": ParamLocalSetting(
            name_setting=NameSetting("pa", "", TPL),
            expr="1.0 - {pb}",
        ),
        "kab": ParamLocalSetting(
            name_setting=NameSetting("kab", "", TPL),
            expr="{kex_ab} * {pb}",
        ),
        "kba": ParamLocalSetting(
            name_setting=NameSetting("kba", "", TPL),
            expr="{kex_ab} * {pa}",
        ),
    }


def register() -> None:
    model_factory.register(name=NAME, setting_maker=make_settings_2st)



from __future__ import annotations

from dataclasses import dataclass
from functools import reduce
from typing import TYPE_CHECKING, Literal

import numpy as np

from chemex.configuration.data import RelaxationDataSettings
from chemex.configuration.experiment import CpmgSettings, ExperimentConfig, ToBeFitted
from chemex.containers.dataset import load_relaxation_dataset
from chemex.experiments.factories import Creators, factories
from chemex.filterers import PlanesFilterer
from chemex.nmr.basis import Basis
from chemex.nmr.liouvillian import LiouvillianIS
from chemex.nmr.spectrometer import Spectrometer
from chemex.plotters.cpmg import CpmgPlotter
from chemex.printers.data import CpmgPrinter

if TYPE_CHECKING:
    from chemex.containers.data import Data
    from chemex.parameters.spin_system import SpinSystem
    from chemex.typing import ArrayBool, ArrayFloat, ArrayInt


EXPERIMENT_NAME = "cpmg_1hn_ap_0013"


class Cpmg1HnAp0013Settings(CpmgSettings):
    name: Literal["cpmg_1hn_ap_0013"]
    time_t2: float
    carrier: float
    pw90: float
    ncyc_max: int
    taua: float = 2.38e-3
    ipap_flg: bool = False
    eburp_flg: bool = False
    reburp_flg: bool = False
    pw_eburp: float = 1.4e-3
    pw_reburp: float = 1.52e-3
    time_equil_1: float = 0.0
    time_equil_2: float = 0.0
    observed_state: Literal["a", "b", "c", "d"] = "a"

    @property
    def t_neg(self) -> float:
        return -2.0 * self.pw90 / np.pi

    @property
    def start(self) -> list[str]:
        return [f"2izsz_{self.observed_state}"]

    @property
    def detection(self) -> str:
        return f"[iz_{self.observed_state}]"


class Cpmg1HnAp0013Config(
    ExperimentConfig[Cpmg1HnAp0013Settings, RelaxationDataSettings],
):
    @property
    def to_be_fitted(self) -> ToBeFitted:
        state = self.experiment.observed_state
        return ToBeFitted(rates=[f"r2_i_{state}"], model_free=[f"tauc_{state}"])


def build_spectrometer(
    config: Cpmg1HnAp0013Config,
    spin_system: SpinSystem,
) -> Spectrometer:
    settings = config.experiment
    conditions = config.conditions

    basis = Basis(type="ixyzsz", spin_system="hn")
    liouvillian = LiouvillianIS(spin_system, basis, conditions)
    spectrometer = Spectrometer(liouvillian)

    spectrometer.carrier_i = settings.carrier
    spectrometer.b1_i = 1 / (4.0 * settings.pw90)
    spectrometer.detection = settings.detection

    return spectrometer


@dataclass
class Cpmg1HnAp0013Sequence:
    settings: Cpmg1HnAp0013Settings

    def _get_delays(
        self,
        ncycs: ArrayFloat,
    ) -> tuple[dict[float, float], dict[float, float], list[float]]:
        ncycs_no_ref = ncycs[ncycs > 0]
        tau_cps = {
            ncyc: self.settings.time_t2 / (4.0 * ncyc) - 0.75 * self.settings.pw90
            for ncyc in ncycs_no_ref
        }
        deltas = {
            ncyc: self.settings.pw90 * (self.settings.ncyc_max - ncyc)
            for ncyc in ncycs_no_ref
        }
        deltas[0.0] = self.settings.pw90 * self.settings.ncyc_max
        delays = [
            self.settings.taua,
            self.settings.t_neg,
            self.settings.pw_eburp,
            0.5 * self.settings.pw_reburp,
            self.settings.time_equil_1,
            self.settings.time_equil_2,
            *tau_cps.values(),
            *deltas.values(),
        ]
        return tau_cps, deltas, delays

    def _get_phases(self, ncyc: ArrayFloat) -> tuple[ArrayInt, ArrayInt]:
        cp_phases1 = np.array(
            [
                [1, 1, 2, 0, 1, 1, 0, 2, 1, 1, 0, 2, 1, 1, 2, 0],
                [2, 0, 3, 3, 0, 2, 3, 3, 0, 2, 3, 3, 2, 0, 3, 3],
            ],
        )
        cp_phases2 = np.array(
            [
                [3, 3, 2, 0, 3, 3, 0, 2, 3, 3, 0, 2, 3, 3, 2, 0],
                [2, 0, 1, 1, 0, 2, 1, 1, 0, 2, 1, 1, 2, 0, 1, 1],
            ],
        )
        indexes = np.arange(int(ncyc))
        phases1 = np.take(cp_phases1, np.flip(indexes), mode="wrap", axis=1)
        phases2 = np.take(cp_phases2, indexes, mode="wrap", axis=1)
        return phases1, phases2

    def calculate(self, spectrometer: Spectrometer, data: Data) -> ArrayFloat:
        ncycs = data.metadata

        # Calculation of the spectrometers corresponding to all the delays
        tau_cps, deltas, all_delays = self._get_delays(ncycs)
        delays = dict(zip(all_delays, spectrometer.delays(all_delays), strict=True))
        d_neg = delays[self.settings.t_neg]
        d_taua = delays[self.settings.taua]
        d_eburp = delays[self.settings.pw_eburp]
        d_reburp = delays[0.5 * self.settings.pw_reburp]
        d_eq_1 = delays[self.settings.time_equil_1]
        d_eq_2 = delays[self.settings.time_equil_2]
        d_delta = {ncyc: delays[delay] for ncyc, delay in deltas.items()}
        d_cp = {ncyc: delays[delay] for ncyc, delay in tau_cps.items()}

        # Calculation of the spectrometers corresponding to all the pulses
        p90 = spectrometer.p90_i
        p180 = spectrometer.p180_i
        pp90_i = spectrometer.perfect90_i
        pp180_isx = spectrometer.perfect180_i[0] @ spectrometer.perfect180_s[0]

        # Calculation of the propagators for INEPT and purge elements
        inept = pp90_i[1] @ d_taua @ pp180_isx @ d_taua @ pp90_i[0]
        zfilter = spectrometer.zfilter

        # Getting the starting magnetization
        start = spectrometer.get_start_magnetization(terms=self.settings.start)

        # Calculating the central refocusing block
        if self.settings.eburp_flg:
            p180pmy = p180[[1, 3]]
            pp90pmy = spectrometer.perfect90_i[[1, 3]]
            e180e_pmy = pp90pmy @ d_eburp @ p180pmy @ d_eburp @ pp90pmy
            middle = np.stack([p180pmy @ e180e_pmy, e180e_pmy @ p180pmy])
        elif self.settings.reburp_flg:
            pp180pmy = spectrometer.perfect180_i[[1, 3]]
            middle = d_reburp @ pp180pmy @ d_reburp
        else:
            middle = p180[[1, 3]]

        # Calculating the intensities as a function of ncyc
        centre = {0.0: d_eq_2 @ d_delta[0] @ p90[0] @ middle @ p90[0] @ d_eq_1}

        for ncyc in set(ncycs) - {0.0}:
            phases1, phases2 = self._get_phases(ncyc)
            echo = d_cp[ncyc] @ p180 @ d_cp[ncyc]
            cpmg1 = reduce(np.matmul, echo[phases1.T])
            cpmg2 = reduce(np.matmul, echo[phases2.T])
            centre[ncyc] = (
                d_eq_2
                @ d_delta[ncyc]
                @ p90[0]
                @ d_neg
                @ cpmg2
                @ middle
                @ cpmg1
                @ d_neg
                @ p90[0]
                @ d_eq_1
            )

        intst = {
            ncyc: spectrometer.detect(inept @ zfilter @ centre[ncyc] @ start)
            for ncyc in set(ncycs)
        }

        if self.settings.ipap_flg:
            intst = {
                ncyc: intst[ncyc]
                + spectrometer.detect(centre[ncyc] @ zfilter @ inept @ start)
                for ncyc in set(ncycs)
            }

        # Return profile
        return np.array([intst[ncyc] for ncyc in ncycs])

    @staticmethod
    def is_reference(metadata: ArrayFloat) -> ArrayBool:
        return metadata == 0


def register() -> None:
    creators = Creators(
        config_creator=Cpmg1HnAp0013Config,
        spectrometer_creator=build_spectrometer,
        sequence_creator=Cpmg1HnAp0013Sequence,
        dataset_creator=load_relaxation_dataset,
        filterer_creator=PlanesFilterer,
        printer_creator=CpmgPrinter,
        plotter_creator=CpmgPlotter,
    )
    factories.register(type=EXPERIMENT_NAME, creators=creators)



from __future__ import annotations

from dataclasses import dataclass
from itertools import combinations, permutations, product
from pathlib import Path
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.cm import get_cmap
from matplotlib.colors import LogNorm
from rich.progress import track

from chemex.messages import print_group_name, print_running_grid
from chemex.optimize.grouping import Group, create_groups
from chemex.optimize.helper import (
    calculate_statistics,
    execute_post_fit,
    execute_post_fit_groups,
    print_header,
    print_values,
)
from chemex.optimize.minimizer import minimize
from chemex.parameters import database

if TYPE_CHECKING:
    from collections.abc import Iterable

    from lmfit.parameter import Parameters

    from chemex.containers.experiments import Experiments
    from chemex.typing import ArrayFloat


@dataclass
class GridResult:
    grid: dict[str, ArrayFloat]
    chisqr: ArrayFloat


def _set_param_values(
    params: Parameters,
    fnames: Iterable[str],
    values: tuple[float, ...],
):
    for fname, value in zip(fnames, values, strict=True):
        params[fname].value = value


def run_group_grid(
    group: Group,
    grid: dict[str, ArrayFloat],
    path: Path,
    fitmethod: str | None,
) -> GridResult:
    group_ids = group.experiments.param_ids
    group_params = database.build_lmfit_params(group_ids)
    group_grid = {
        param_id: values for param_id, values in grid.items() if param_id in group_ids
    }

    grid_ids = tuple(group_grid)
    shape = tuple(len(values) for values in group_grid.values())
    grid_size = np.prod(shape)

    basename = group.path if group.path != Path() else Path("grid")
    filename = path / "Grid" / f"{basename}.out"
    filename.parent.mkdir(parents=True, exist_ok=True)

    best_chisqr = np.inf
    best_params = group_params

    with filename.open("w") as fileout:
        fileout.write(print_header(group_grid))

        chisqr_list: list[float] = []

        grid_values = product(*group_grid.values())

        for values in track(grid_values, total=float(grid_size), description="   "):
            _set_param_values(group_params, grid_ids, values)
            optimized_params = minimize(group.experiments, group_params, fitmethod)
            stats = calculate_statistics(group.experiments, optimized_params)
            chisqr: float = stats.get("chisqr", np.inf)
            chisqr_list.append(chisqr)
            fileout.write(print_values(values, chisqr))
            fileout.flush()

            if chisqr < best_chisqr:
                best_chisqr = chisqr
                best_params = optimized_params

    chisqr_array = np.array(chisqr_list).reshape(shape)
    database.update_from_parameters(best_params)

    return GridResult(group_grid, chisqr_array)


def _reshape_chisqr(
    grid_ref: dict[str, ArrayFloat],
    grid_result: GridResult,
) -> ArrayFloat:
    keys = list(grid_result.grid)
    order = [keys.index(key) for key in grid_ref if key in keys]
    axes_to_reduce = tuple(sorted(set(range(len(keys))) - set(order)))
    order.extend(axes_to_reduce)

    # After transpose, axes are shuffled
    axes_to_reduce = tuple(order.index(index) for index in axes_to_reduce)

    chisqr_final = grid_result.chisqr.transpose(order)
    chisqr_final = np.minimum.reduce(chisqr_final, axis=axes_to_reduce)

    shape = tuple(
        len(grid_ref[key]) if key in grid_result.grid else 1 for key in grid_ref
    )
    return chisqr_final.reshape(shape)


def _get_grids(
    grid: dict[str, ArrayFloat],
    grid_results: list[GridResult],
) -> list[dict[str, ArrayFloat]]:
    grid_params = {tuple(sorted(grid_result.grid)) for grid_result in grid_results}
    grid_params_tmp = grid_params.copy()
    for params1, params2 in permutations(grid_params, 2):
        if set(params1) <= set(params2):
            grid_params_tmp.remove(params1)
    return [{key: grid[key] for key in params} for params in grid_params_tmp]


def combine_grids(
    grid: dict[str, ArrayFloat],
    grid_results: list[GridResult],
) -> list[GridResult]:
    grids = _get_grids(grid, grid_results)

    results: list[GridResult] = []

    for grid_ref in grids:
        shape = tuple(len(values) for values in grid_ref.values())

        chisqr_sum = np.zeros(shape)

        for grid_result in grid_results:
            chisqr_final = _reshape_chisqr(grid_ref, grid_result)
            chisqr_sum += chisqr_final

        result = GridResult(grid_ref, chisqr_sum)
        results.append(result)

    return results


def set_params_from_grid(grids_1d: Iterable[GridResult]):
    par_values = {}
    for grid_result in grids_1d:
        id_, values = next(iter(grid_result.grid.items()))
        par_values[id_] = values[grid_result.chisqr.argmin()]
    database.set_param_values(par_values)


def make_grids_nd(
    grid: dict[str, ArrayFloat],
    grids_combined: list[GridResult],
    ndim: int,
) -> list[GridResult]:
    grids: list[GridResult] = []
    parameters = database.get_parameters(grid)
    ids = sorted(grid, key=lambda x: parameters[x].param_name)
    for selection in combinations(ids, ndim):
        for grid_result in grids_combined:
            if set(selection) <= set(grid_result.grid):
                grid_ref = {fname: grid[fname] for fname in selection}
                chisqr_final = _reshape_chisqr(grid_ref, grid_result)
                grids.append(GridResult(grid_ref, chisqr_final))
                break
    return grids


def plot_grid_1d(grids_1d: list[GridResult], path: Path) -> None:
    """Visualize the result of the brute force grid search.

    The output file will display the chi-square values per parameter.

    """
    if not grids_1d:
        return

    with PdfPages(str(path / "grid_1d.pdf")) as pdf:
        for grid_result in grids_1d:
            parameters = database.get_parameters(grid_result.grid)
            ((id_, values),) = list(grid_result.grid.items())
            _fig, ax = plt.subplots(1, 1)
            ax.plot(values, grid_result.chisqr, "o", ms=3)
            best_value = parameters[id_].value
            if best_value is not None:
                ax.axvline(best_value, ls="dashed", color=(0.5, 0.5, 0.5))
            ax.set_xlabel(str(parameters[id_].param_name))
            ax.set_ylabel(r"$\chi^{2}$")
            pdf.savefig()
            plt.close()


def plot_grid_2d(grids_2d: list[GridResult], path: Path) -> None:
    """Visualize the result of the brute force grid search.

    The output file will display the chi-square contour
    plots for all combination of two parameters.

    """
    if not grids_2d:
        return

    with PdfPages(str(path / "grid_2d.pdf")) as pdf:
        for grid_result in grids_2d:
            parameters = database.get_parameters(grid_result.grid)
            (id_x, values_x), (id_y, values_y) = grid_result.grid.items()
            fig, ax = plt.subplots(1, 1)
            grid_x, grid_y = np.meshgrid(values_x, values_y)
            best_value_x = parameters[id_x].value
            best_value_y = parameters[id_y].value
            if best_value_x is not None:
                ax.axvline(best_value_x, ls="dashed", color=(0.5, 0.5, 0.5), zorder=-1)
            if best_value_y is not None:
                ax.axhline(best_value_y, ls="dashed", color=(0.5, 0.5, 0.5), zorder=-1)
            cs = ax.scatter(
                grid_x,
                grid_y,
                c=grid_result.chisqr.T,
                norm=LogNorm(),
                cmap=get_cmap("viridis_r"),
            )
            cbar = fig.colorbar(cs)
            cbar.set_label(r"$\chi^{2}$")
            ax.set_xlabel(str(parameters[id_x].param_name))
            ax.set_ylabel(str(parameters[id_y].param_name))
            pdf.savefig()
            plt.close()


def run_grid(
    experiments: Experiments,
    grid_raw: list[str],
    path: Path,
    plot: str,
    fitmethod: str,
) -> None:
    print_running_grid()

    grid = database.parse_grid(grid_raw)

    groups = create_groups(experiments)

    grid_results: list[GridResult] = []
    for group in groups:
        if message := group.message:
            print_group_name(message)
        grid_result = run_group_grid(group, grid, path, fitmethod)
        grid_results.append(grid_result)

    grids_combined = combine_grids(grid, grid_results)
    grids_1d = make_grids_nd(grid, grids_combined, 1)
    grids_2d = make_grids_nd(grid, grids_combined, 2)
    set_params_from_grid(grids_1d)
    plot_grid_1d(grids_1d, path / "Grid")
    plot_grid_2d(grids_2d, path / "Grid")

    if len(groups) > 1:
        execute_post_fit_groups(experiments, path, plot)
    else:
        execute_post_fit(experiments, path, plot != "nothing")



from __future__ import annotations

from typing import TYPE_CHECKING

from chemex.models.constraints import pop_2st
from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting
from chemex.parameters.userfunctions import user_function_registry

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions

NAME = "2st_hd"


def make_settings_2st_hd(conditions: Conditions) -> dict[str, ParamLocalSetting]:
    d2o = conditions.d2o if conditions.d2o is not None else 0.1
    return {
        "d2o": ParamLocalSetting(
            name_setting=NameSetting("d2o", "", ("d2o",)),
            value=d2o,
            min=0.0,
            max=1.0,
        ),
        "kdh": ParamLocalSetting(
            name_setting=NameSetting("kdh", "g", ("temperature",)),
            value=1.0,
            min=0.0,
            vary=True,
        ),
        "phi": ParamLocalSetting(
            name_setting=NameSetting("phi", "g", ("temperature",)),
            value=1.1,
            min=0.75,
            max=1.50,
        ),
        "kab": ParamLocalSetting(
            name_setting=NameSetting("kab", "g", ("temperature", "d2o")),
            min=0.0,
            expr="{d2o} * {kdh} * {phi}",
        ),
        "kba": ParamLocalSetting(
            name_setting=NameSetting("kba", "g", ("temperature", "d2o")),
            expr="(1.0 - {d2o}) * {kdh}",
        ),
        "pa": ParamLocalSetting(
            name_setting=NameSetting("pa", "g", ("temperature", "d2o")),
            expr="pop_2st({kab}, {kba})['pa']",
        ),
        "pb": ParamLocalSetting(
            name_setting=NameSetting("pb", "g", ("temperature", "d2o")),
            expr="pop_2st({kab}, {kba})['pb']",
        ),
    }


def register() -> None:
    model_factory.register(name=NAME, setting_maker=make_settings_2st_hd)
    user_function_registry.register(name=NAME, user_functions={"pop_2st": pop_2st})



from collections.abc import Iterable
from itertools import chain, combinations
from typing import TypeVar

T = TypeVar("T")


def powerset(iterable: Iterable[T]) -> Iterable[tuple[T, ...]]:
    """Generates all possible non-empty subsets of the given iterable.

    This function returns the power set of the input iterable, excluding the empty set.
    The subsets are returned as tuples. The order of subsets follows the order of
    elements in the input iterable.

    Args:
        iterable (Iterable[Any]): An iterable for which to generate the power set.

    Returns:
        Iterable[Any]: An iterable of tuples, each representing a non-empty subset of
        the input iterable.

    Examples:
        >>> powerset([1, 2, 3])
        (1,) (2,) (3,) (1, 2) (1, 3) (2, 3) (1, 2, 3)

    Note:
        This implementation uses `itertools.chain` and `itertools.combinations`, and is
        efficient for small to medium-sized iterables.

    """
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(1, len(s) + 1))



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal

import numpy as np

from chemex.configuration.data import CestDataSettings
from chemex.configuration.experiment import CestSettings, ExperimentConfig, ToBeFitted
from chemex.containers.dataset import load_relaxation_dataset
from chemex.experiments.factories import Creators, factories
from chemex.filterers import CestFilterer
from chemex.nmr.basis import Basis
from chemex.nmr.liouvillian import LiouvillianIS
from chemex.nmr.spectrometer import Spectrometer
from chemex.plotters.cest import CestPlotter
from chemex.printers.data import CestPrinter

if TYPE_CHECKING:
    from chemex.containers.data import Data
    from chemex.parameters.spin_system import SpinSystem
    from chemex.typing import ArrayBool, ArrayFloat


EXPERIMENT_NAME = "cest_1hn_ap"

OFFSET_REF = 1e4


class Cest1HnApSettings(CestSettings):
    name: Literal["cest_1hn_ap"]
    time_t1: float
    carrier: float
    b1_frq: float
    b1_inh_scale: float = 0.1
    b1_inh_res: int = 11
    observed_state: Literal["a", "b", "c", "d"] = "a"

    @property
    def start(self) -> list[str]:
        return [f"2izsz_{self.observed_state}"]

    @property
    def detection(self) -> str:
        return f"[2izsz_{self.observed_state}]"


class Cest1HnApConfig(ExperimentConfig[Cest1HnApSettings, CestDataSettings]):
    @property
    def to_be_fitted(self) -> ToBeFitted:
        state = self.experiment.observed_state
        return ToBeFitted(
            rates=["r2_i", f"r1_i_{state}", f"r1_s_{state}", f"etaxy_i_{state}"],
            model_free=[f"tauc_{state}", f"s2_{state}", f"khh_{state}"],
        )


def build_spectrometer(
    config: Cest1HnApConfig,
    spin_system: SpinSystem,
) -> Spectrometer:
    settings = config.experiment
    conditions = config.conditions

    basis = Basis(type="ixyzsz", spin_system="hn")
    liouvillian = LiouvillianIS(spin_system, basis, conditions)
    spectrometer = Spectrometer(liouvillian)

    spectrometer.carrier_i = settings.carrier
    spectrometer.b1_i = settings.b1_frq
    spectrometer.b1_i_inh_scale = settings.b1_inh_scale
    spectrometer.b1_i_inh_res = settings.b1_inh_res

    spectrometer.detection = settings.detection

    return spectrometer


@dataclass
class Cest1HnApSequence:
    settings: Cest1HnApSettings

    @staticmethod
    def is_reference(metadata: ArrayFloat) -> ArrayBool:
        return np.abs(metadata) > OFFSET_REF

    def calculate(self, spectrometer: Spectrometer, data: Data) -> ArrayFloat:
        offsets = data.metadata

        start = spectrometer.get_start_magnetization(terms=self.settings.start)

        intensities: dict[float, ArrayFloat] = {}

        for offset in set(offsets):
            intensities[offset] = start

            if self.is_reference(offset):
                continue

            spectrometer.offset_i = offset

            intensities[offset] = (
                spectrometer.pulse_i(self.settings.time_t1, 0.0) @ intensities[offset]
            )

        return np.array(
            [spectrometer.detect(intensities[offset]) for offset in offsets],
        )


def register() -> None:
    creators = Creators(
        config_creator=Cest1HnApConfig,
        spectrometer_creator=build_spectrometer,
        sequence_creator=Cest1HnApSequence,
        dataset_creator=load_relaxation_dataset,
        filterer_creator=CestFilterer,
        printer_creator=CestPrinter,
        plotter_creator=CestPlotter,
    )
    factories.register(type=EXPERIMENT_NAME, creators=creators)



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from re import compile
from typing import TYPE_CHECKING

from chemex.parameters import database
from chemex.parameters.name import ParamName
from chemex.parameters.setting import ParamSetting

if TYPE_CHECKING:
    from pathlib import Path

    from chemex.containers.experiments import Experiments

Parameters = dict[ParamName, ParamSetting]

RE_GROUPNAME = compile(r"^[A-Za-z0-9_-]+$")


@dataclass
class GlobalLocalParameters:
    global_: Parameters
    local: Parameters

    def __bool__(self) -> bool:
        return bool(self.global_) or bool(self.local)


@dataclass
class ClassifiedParameters:
    fitted: GlobalLocalParameters
    fixed: GlobalLocalParameters
    constrained: GlobalLocalParameters


def _format_fitted(param: ParamSetting) -> str:
    error = f"±{param.stderr:.5e}" if param.stderr else "(error not calculated)"
    return f"{param.value: .5e} # {error}"


def _format_constrained(param: ParamSetting) -> str:
    if param.value is None:
        return ""

    error = f"±{param.stderr:.5e} " if param.stderr else ""
    constraint = param.expr
    parameters = database.get_parameters(param.dependencies)
    for param_id, parameter in parameters.items():
        constraint = constraint.replace(param_id, str(parameter.param_name))
    return f"{param.value: .5e} # {error}({constraint})"


def _format_fixed(param: ParamSetting) -> str:
    return f"{param.value: .5e} # (fixed)"


_format_param = {
    "fitted": _format_fitted,
    "constrained": _format_constrained,
    "fixed": _format_fixed,
}


def _params_to_strings(
    parameters: GlobalLocalParameters,
    status: str,
) -> dict[str, dict[str, str]]:
    result: defaultdict[str, dict[str, str]] = defaultdict(dict)

    for pname, param in parameters.global_.items():
        result["GLOBAL"][pname.section_res] = _format_param[status](param)

    for pname, param in parameters.local.items():
        result[pname.section][str(pname.spin_system)] = _format_param[status](param)

    return result


def _quote(text: str) -> str:
    text = text.strip(" ,")
    return text if RE_GROUPNAME.match(text) else f'"{text}"'


def _format_strings(par_strings: dict[str, dict[str, str]]) -> str:
    result: list[str] = []
    for section, key_values in par_strings.items():
        result.append(f"[{_quote(section)}]")
        width = len(max(key_values, key=len))
        result.extend(
            f"{_quote(key):<{width}} = {value}" for key, value in key_values.items()
        )
        result.append("")
    return "\n".join(result)


def write_file(parameters: GlobalLocalParameters, status: str, path: Path) -> None:
    if not parameters:
        return

    par_strings = _params_to_strings(parameters, status)
    formatted_strings = _format_strings(par_strings)
    filename = path / f"{status}.toml"
    filename.write_text(formatted_strings)


def classify_global(parameters: Parameters) -> GlobalLocalParameters:
    local = {}
    global_ = {}

    for pname, parameter in parameters.items():
        if parameter.param_name.spin_system:
            local[pname] = parameter
        else:
            global_[pname] = parameter

    return GlobalLocalParameters(global_, local)


def classify_parameters(experiments: Experiments) -> ClassifiedParameters:
    param_ids = experiments.param_ids
    parameters = {
        param.param_name: param for param in database.get_parameters(param_ids).values()
    }

    constrained = {pname: param for pname, param in parameters.items() if param.expr}
    fitted = {
        pname: param
        for pname, param in parameters.items()
        if param.vary and pname not in constrained
    }

    fixed_ids = set(parameters) - set(fitted) - set(constrained)
    fixed = {
        pname: parameter
        for pname, parameter in parameters.items()
        if pname in fixed_ids
    }

    return ClassifiedParameters(
        classify_global(fitted),
        classify_global(fixed),
        classify_global(constrained),
    )


def write_parameters(experiments: Experiments, path: Path):
    """Write the model parameter values and their uncertainties to a file."""
    path_par = path / "Parameters"
    path_par.mkdir(parents=True, exist_ok=True)
    classified_parameters = classify_parameters(experiments)

    write_file(classified_parameters.fitted, "fitted", path_par)
    write_file(classified_parameters.fixed, "fixed", path_par)
    write_file(classified_parameters.constrained, "constrained", path_par)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from matplotlib.figure import Figure

if TYPE_CHECKING:
    from matplotlib.axes import Axes

    from chemex.containers.data import Data
    from chemex.typing import ArrayFloat

_RED100 = "#FFCDD2"
_RED300 = "#E57373"
_RED700 = "#D32F2F"

LARGE_VALUE = 1e16


def _create_fig(name: str) -> tuple[Figure, Axes, Axes]:
    fig = Figure()
    ax1, ax2 = fig.subplots(2, 1, sharex="all", gridspec_kw={"height_ratios": [1, 4]})
    fig.align_labels()
    fig.suptitle(f"{name.upper()}")
    ax1.set_ylabel("Residuals")
    ax1.ticklabel_format(style="sci", scilimits=(0, 0), axis="y", useMathText=True)

    return fig, ax1, ax2


def get_grid(values: ArrayFloat, size: int = 400, extension: float = 0.0) -> ArrayFloat:
    value_min = np.min(values)
    value_max = np.max(values)
    extra = (value_max - value_min) * extension
    return np.linspace(value_min - extra, value_max + extra, size)


def _plot_fit(data_calc: Data, ax2: Axes):
    fit_x = data_calc.metadata
    fit_y = data_calc.calc
    range_x = get_grid(fit_x, 2, 0.02)
    ax2.set_xlim(*range_x)
    ax2.plot(fit_x, fit_y, linestyle="-", color=_RED300)


def _plot_exp(data_exp: Data, ax1: Axes, ax2: Axes):
    exp_x = data_exp.metadata
    exp_y = data_exp.exp
    exp_e = abs(data_exp.err)
    res_y = data_exp.exp - data_exp.calc

    m_sel = data_exp.mask
    m_inf = exp_e.sum(axis=1) > LARGE_VALUE

    s1 = m_sel & ~m_inf
    ax1.errorbar(exp_x[s1], res_y[s1], exp_e[s1].T, fmt=".", color=_RED700, zorder=3)
    ax2.errorbar(exp_x[s1], exp_y[s1], exp_e[s1].T, fmt=".", color=_RED700, zorder=3)

    s3 = ~m_sel
    ax1.errorbar(exp_x[s3], res_y[s3], exp_e[s3].T, fmt=".", color=_RED100, zorder=3)
    ax2.errorbar(exp_x[s3], exp_y[s3], exp_e[s3].T, fmt=".", color=_RED100, zorder=3)

    range1_y = ax1.get_ylim()
    range2_y = ax2.get_ylim()

    exp_e[exp_e == np.inf] = 500.0
    s2 = m_sel & m_inf
    ax1.errorbar(exp_x[s2], res_y[s2], exp_e[s2].T, fmt=".", color=_RED700, zorder=3)
    ax2.errorbar(exp_x[s2], exp_y[s2], exp_e[s2].T, fmt=".", color=_RED700, zorder=3)

    ax1.set_ylim(*range1_y)
    ax2.set_ylim(*range2_y)


def plot_profile(name: str, data_exp: Data, data_calc: Data) -> Figure:
    fig, ax1, ax2 = _create_fig(name)

    _plot_fit(data_calc, ax2)

    if data_exp.size:
        _plot_exp(data_exp, ax1, ax2)

    for axis in (ax1, ax2):
        axis.axhline(0, color="k", linewidth=0.5, zorder=1)

    return fig



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np
from scipy.optimize import root

from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting
from chemex.parameters.userfunctions import user_function_registry

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions
    from chemex.typing import ArrayFloat


NAME = "4st_binding_partner_2st"

TPL = ("temperature", "p_total", "l_total")


def calculate_residuals(
    concentrations: ArrayFloat,
    p_total: float,
    l_total: float,
    kd1: float,
    kd2: float,
    keq_l: float,
    keq_pl: float,
) -> ArrayFloat:
    p, l1, l2, pl1, pl2, pl3 = concentrations
    return np.array(
        [
            l_total - (l1 + l2 + pl1 + pl2 + pl3),
            p_total - (p + pl1 + pl2 + pl3),
            kd1 * pl1 - p * l1,
            kd2 * pl2 - p * l2,
            keq_l * l1 - l2,
            keq_pl * pl2 - pl3,
        ],
    )


@lru_cache(maxsize=100)
def calculate_concentrations(
    p_total: float,
    l_total: float,
    kd1: float,
    kd2: float,
    keq_l: float,
    keq_pl: float,
) -> dict[str, float]:
    concentrations_start = (p_total, l_total / 2, l_total / 2, 0.0, 0.0, 0.0)
    results = root(
        calculate_residuals,
        concentrations_start,
        args=(p_total, l_total, kd1, kd2, keq_l, keq_pl),
    )
    return dict(zip(("p", "l1", "l2", "pl1", "pl2", "pl3"), results["x"], strict=True))


def make_settings_4st_binding_partner_2st(
    conditions: Conditions,
) -> dict[str, ParamLocalSetting]:
    p_total = conditions.p_total
    l_total = conditions.l_total
    if p_total is None:
        msg = f"'p_total' must be specified to use the '{NAME}' model"
        raise ValueError(msg)
    if l_total is None:
        msg = f"'l_total' must be specified to use the '{NAME}' model"
        raise ValueError(msg)
    return {
        "koff_ab": ParamLocalSetting(
            name_setting=NameSetting("koff_ab", "", ("temperature",)),
            value=100.0,
            min=0.0,
            vary=True,
        ),
        "kd_ab": ParamLocalSetting(
            name_setting=NameSetting("kd_ab", "", ("temperature",)),
            value=1e-3,
            min=0.0,
            vary=True,
        ),
        "koff_ac": ParamLocalSetting(
            name_setting=NameSetting("koff_ac", "", ("temperature",)),
            value=100.0,
            min=0.0,
            vary=True,
        ),
        "kd_ac": ParamLocalSetting(
            name_setting=NameSetting("kd_ac", "", ("temperature",)),
            value=1e-3,
            min=0.0,
            vary=True,
        ),
        "keq_l": ParamLocalSetting(
            name_setting=NameSetting("keq_l", "", ("temperature",)),
            value=1.0,
            min=0.0,
            vary=True,
        ),
        "keq_pl": ParamLocalSetting(
            name_setting=NameSetting("keq_pl", "", ("temperature",)),
            value=1.0,
            min=0.0,
            vary=True,
        ),
        "kex_bc": ParamLocalSetting(
            name_setting=NameSetting("kex_bc", "", ("temperature",)),
            value=1e3,
            min=0.0,
            vary=True,
        ),
        "kex_cd": ParamLocalSetting(
            name_setting=NameSetting("kex_cd", "", ("temperature",)),
            value=1e3,
            min=0.0,
            vary=True,
        ),
        "kon_ab": ParamLocalSetting(
            name_setting=NameSetting("kon_ab", "", ("temperature",)),
            expr="{koff_ab} / max({kd_ab}, 1e-100)",
        ),
        "kon_ac": ParamLocalSetting(
            name_setting=NameSetting("kon_ac", "", ("temperature",)),
            expr="{koff_ac} / max({kd_ac}, 1e-100)",
        ),
        "p_free": ParamLocalSetting(
            name_setting=NameSetting("p_free", "", TPL),
            expr=f"calc_conc({p_total},{l_total},{{kd_ab}},{{kd_ac}},{{keq_l}},{{keq_pl}})['p']",
        ),
        "l1_free": ParamLocalSetting(
            name_setting=NameSetting("l1_free", "", TPL),
            expr=f"calc_conc({p_total},{l_total},{{kd_ab}},{{kd_ac}},{{keq_l}},{{keq_pl}})['l1']",
        ),
        "l2_free": ParamLocalSetting(
            name_setting=NameSetting("l1_free", "", TPL),
            expr=f"calc_conc({p_total},{l_total},{{kd_ab}},{{kd_ac}},{{keq_l}},{{keq_pl}})['l2']",
        ),
        "pl1": ParamLocalSetting(
            name_setting=NameSetting("pl1", "", TPL),
            expr=f"calc_conc({p_total},{l_total},{{kd_ab}},{{kd_ac}},{{keq_l}},{{keq_pl}})['pl1']",
        ),
        "pl2": ParamLocalSetting(
            name_setting=NameSetting("pl2", "", TPL),
            expr=f"calc_conc({p_total},{l_total},{{kd_ab}},{{kd_ac}},{{keq_l}},{{keq_pl}})['pl2']",
        ),
        "pl3": ParamLocalSetting(
            name_setting=NameSetting("pl3", "", TPL),
            expr=f"calc_conc({p_total},{l_total},{{kd_ab}},{{kd_ac}},{{keq_l}},{{keq_pl}})['pl3']",
        ),
        "kab": ParamLocalSetting(
            name_setting=NameSetting("kab", "", TPL),
            expr="{kon_ab} * {l1_free}",
        ),
        "kba": ParamLocalSetting(
            name_setting=NameSetting("kba", "", TPL),
            expr="{koff_ab}",
        ),
        "kac": ParamLocalSetting(
            name_setting=NameSetting("kac", "", TPL),
            expr="{kon_ac} * {l2_free}",
        ),
        "kca": ParamLocalSetting(
            name_setting=NameSetting("kca", "", TPL),
            expr="{koff_ac}",
        ),
        "kbc": ParamLocalSetting(
            name_setting=NameSetting("kbc", "", TPL),
            expr="{kex_bc} * {pl2} / max({pl1} + {pl2}, 1e-100)",
        ),
        "kcb": ParamLocalSetting(
            name_setting=NameSetting("kcb", "", TPL),
            expr="{kex_bc} * {pl1} / max({pl1} + {pl2}, 1e-100)",
        ),
        "kcd": ParamLocalSetting(
            name_setting=NameSetting("kcd", "", TPL),
            expr="{kex_cd} * {pl3} / max({pl2} + {pl3}, 1e-100)",
        ),
        "kdc": ParamLocalSetting(
            name_setting=NameSetting("kdc", "", TPL),
            expr="{kex_cd} * {pl2} / max({pl2} + {pl3}, 1e-100)",
        ),
        "pa": ParamLocalSetting(
            name_setting=NameSetting("pa", "", TPL),
            expr=f"{{p_free}} / {p_total}",
        ),
        "pb": ParamLocalSetting(
            name_setting=NameSetting("pb", "", TPL),
            expr=f"{{pl1}} / {p_total}",
        ),
        "pc": ParamLocalSetting(
            name_setting=NameSetting("pc", "", TPL),
            expr=f"{{pl2}} / {p_total}",
        ),
        "pd": ParamLocalSetting(
            name_setting=NameSetting("pc", "", TPL),
            expr=f"{{pl3}} / {p_total}",
        ),
    }


def register() -> None:
    model_factory.register(
        name=NAME,
        setting_maker=make_settings_4st_binding_partner_2st,
    )
    user_functions = {
        "calc_conc": calculate_concentrations,
    }
    user_function_registry.register(name=NAME, user_functions=user_functions)



from __future__ import annotations

import sys
from collections.abc import Iterable, MutableMapping
from typing import TYPE_CHECKING, Any

import tomli as tomllib  # Will be part of standard library in Python 3.11

from chemex.messages import print_file_not_found, print_toml_error

if TYPE_CHECKING:
    from pathlib import Path


def read_toml(filename: Path) -> MutableMapping[str, Any]:
    """Read and parse the experiment configuration file with 'toml."""
    with filename.open(mode="rb") as file:
        try:
            config = tomllib.load(file)
        except FileNotFoundError:
            print_file_not_found(filename)
            sys.exit(1)
        except (tomllib.TOMLDecodeError, TypeError) as error:
            print_toml_error(filename, error)
            sys.exit(1)

        else:
            return config

    # Recursive dictionary merge
    # Copyright (C) 2016 Paul Durivage <pauldurivage+github@gmail.com>
    #
    # This program is free software: you can redistribute it and/or modify
    # it under the terms of the GNU General Public License as published by
    # the Free Software Foundation, either version 3 of the License, or
    # (at your option) any later version.
    #
    # This program is distributed in the hope that it will be useful,
    # but WITHOUT ANY WARRANTY; without even the implied warranty of
    # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    # GNU General Public License for more details.
    #
    # You should have received a copy of the GNU General Public License
    # along with this program.  If not, see <https://www.gnu.org/licenses/>.


def _dict_merge(
    dct: MutableMapping[str, Any], merge_dct: MutableMapping[str, Any]
) -> None:
    """Recursive dictionary merge.

    Inspired by :meth:``dict.update()``, instead of updating only top-level
    keys, dict_merge recurses down into dicts nested to an arbitrary depth,
    updating keys. The ``merge_dct`` is merged into ``dct``.

    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None

    """
    for key in merge_dct:
        if (
            key in dct
            and isinstance(dct[key], dict)
            and isinstance(merge_dct[key], MutableMapping)
        ):
            _dict_merge(dct[key], merge_dct[key])
        else:
            dct[key] = merge_dct[key]


def read_toml_multi(filenames: Iterable[Path]) -> MutableMapping[str, Any]:
    """Read and parse multiple experiment configuration files with 'toml'."""
    config: dict[str, Any] = {}
    for filename in filenames:
        _dict_merge(config, read_toml(filename))
    return config


def normalize_path(working_dir: Path, filename: Path) -> Path:
    """Normalize the path of a filename relative to a specific directory."""
    path = filename
    if not path.is_absolute():
        path = working_dir / path
    return path.resolve()



"""A simple plugin loader."""
from __future__ import annotations

import importlib
from pkgutil import iter_modules
from typing import Protocol, cast

from chemex.models import kinetic


class SettingModuleInterface(Protocol):
    """Represents a setting module interface.
    A setting module plugin has a single register function.
    """

    @staticmethod
    def register() -> None:
        """Register the necessary items related to the settings."""
        ...


def import_module(name: str) -> SettingModuleInterface:
    """Imports a module given a name."""
    return cast(SettingModuleInterface, importlib.import_module(name))


def register_kinetic_settings() -> None:
    """Loads the plugins defined in the plugins list."""
    for module in iter_modules(kinetic.__path__):
        module_name = f"{kinetic.__name__}.{module.name}"
        setting_module = import_module(module_name)
        setting_module.register()



from __future__ import annotations

from functools import cache, cached_property, total_ordering
from typing import TYPE_CHECKING, Annotated, Any, Literal, TypeVar

from pydantic import PlainValidator

from .atom import Atom
from .group import Group
from .spin import Spin
from .utilities import powerset

if TYPE_CHECKING:
    from collections.abc import Hashable, Iterable, Sequence

    from chemex.nmr.basis import Basis

    from .nucleus import Nucleus


Self = TypeVar("Self", bound="SpinSystem")

SPIN_ALIASES = "isx"


@total_ordering
class SpinSystem:
    @staticmethod
    @cache
    def _parse_spin_system(name: str) -> dict[str, Spin]:
        if not name:
            return {}
        split = name.split("-")
        spins: dict[str, Spin] = {}
        last_group = None
        for short_name, name_spin in zip(SPIN_ALIASES, split, strict=False):
            spin = Spin(name_spin, last_group)
            spins[short_name] = spin
            last_group = spin.group
        return spins

    @staticmethod
    def _spins2name(spins: Iterable[Spin]) -> str:
        spin_names: list[str] = []
        last_group: Group = Group("")
        for spin in spins:
            spin_name = str(spin.atom) if spin.group == last_group else str(spin)
            spin_names.append(spin_name)
            last_group = spin.group
        return "-".join(spin_names)

    def __init__(self, name: str | int | None = None) -> None:
        if name is None:
            name = ""
        if isinstance(name, int):
            name = str(name)
        self.spins = self._parse_spin_system(name.strip().upper())
        self.name = self._spins2name(self.spins.values())
        self.search_keys: set[Hashable] = set()
        self.search_keys = self.search_keys.union(
            *(spin.search_keys for spin in self.spins.values()),
        )

    def __deepcopy__(self: Self, memo: dict[Any, Any]) -> Self:
        return type(self)(self.name)

    @cached_property
    def names(self) -> dict[str, str]:
        result: dict[str, str] = {}
        for alias_set in powerset(SPIN_ALIASES):
            if set(alias_set).issubset(self.spins):
                key = "".join(alias_set)
                name = self._spins2name(self.spins[alias] for alias in alias_set)
                result[key] = name
        return result

    @cached_property
    def groups(self) -> dict[str, Group]:
        return {alias: spin.group for alias, spin in self.spins.items()}

    @cached_property
    def symbols(self) -> dict[str, str]:
        return {alias: group.symbol for alias, group in self.groups.items()}

    @cached_property
    def numbers(self) -> dict[str, int]:
        return {alias: group.number for alias, group in self.groups.items()}

    @cached_property
    def atoms(self) -> dict[str, Atom]:
        return {alias: spin.atom for alias, spin in self.spins.items()}

    @cached_property
    def nuclei(self) -> dict[str, Nucleus]:
        return {alias: atom.nucleus for alias, atom in self.atoms.items()}

    def match(self: Self, other: Self) -> bool:
        spin_pairs = zip(self.spins.values(), other.spins.values(), strict=False)
        return all(spin.match(other_spin) for spin, other_spin in spin_pairs)

    def part_of(self: Self, selection: Sequence[Self] | str) -> bool:
        if isinstance(selection, str):
            return selection.lower() in ("all", "*")
        return any(item.match(self) for item in selection)

    def correct(self: Self, basis: Basis) -> Self:
        spins: list[Spin] = []
        last_spin = Spin("")
        for letter, atom in basis.atoms.items():
            spin = Spin(self.spins.get(letter, last_spin).name)
            if not spin.atom.name.startswith(atom.upper()):
                spin.atom = Atom(f"{atom}{spin.atom.name[1:]}")
            last_spin = spin
            spins.append(spin)
        return type(self)(self._spins2name(spins))

    def build_sub_spin_system(
        self: Self,
        spin_system_part: Literal["i", "s", "is", "g", ""],
    ) -> Self:
        if spin_system_part == "g":
            return type(self)(str(self.groups.get("i", "")))
        if spin_system_part == "i":
            return type(self)(str(self.spins.get("i", "")))
        if spin_system_part == "s":
            return type(self)(str(self.spins.get("s", "")))
        if spin_system_part == "is":
            return self
        return type(self)()

    def __and__(self: Self, other: object) -> Self:
        if not isinstance(other, type(self)):
            return NotImplemented
        if self == other:
            return type(self)(self.name)
        if spins := set(self.spins.values()) & set(other.spins.values()):
            return type(self)("-".join(spin.name for spin in spins))
        groups = set(self.groups.values()) & set(other.groups.values())
        if len(groups) == 1:
            return type(self)("-".join(group.name for group in groups))
        return type(self)("")

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return tuple(self.spins.values()) < tuple(other.spins.values())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name)

    def __bool__(self) -> bool:
        return bool(self.name)

    def __str__(self) -> str:
        return self.name


# We now create an `Annotated` wrapper that we'll use as the annotation for
# fields on `BaseModel`
PydanticSpinSystem = Annotated[SpinSystem, PlainValidator(lambda x: SpinSystem(x))]


# if __name__ == "__main__":
#     print(SpinSystem("G23N-G23HN"), SpinSystem("GLY023N-HN"), SpinSystem("g23N-hN"))
#     print(SpinSystem("G23N-G23HN") == SpinSystem("GLY023N-HN"))
#     print(SpinSystem("G23N-G23HN").match(SpinSystem("GLY023N-HN")))
#     print(SpinSystem("G23N-G23HN") & SpinSystem("G23C"))
#     print(SpinSystem("") == SpinSystem())
#     group = Group("L99")
#     spin = Spin("HD1", group)
#     print(f"spin = {spin}, spin.group = {spin.group}, spin.atom = {spin.atom}")
#     print(SpinSystem("GLY023N-HN").search_keys)
#     print(SpinSystem("GLY023N-HN").names)
#     print(SpinSystem("G23C1'").names)



from __future__ import annotations

import sys
from dataclasses import dataclass
from string import ascii_lowercase

from chemex.messages import print_model_error
from chemex.models.factory import model_factory


@dataclass
class _Model:
    _name: str = "2st"
    _states: str = "ab"
    _model_free: bool = False
    _temp_coef: bool = False

    @staticmethod
    def validate_model_name(name: str) -> str:
        if name not in model_factory.set:
            print_model_error(name)
            sys.exit()
        return name

    def set_model(self, name: str) -> None:
        kinetic_model_name, *ext = name.split(".")
        self._name = self.validate_model_name(kinetic_model_name)
        state_nb = int(kinetic_model_name[0])
        self._states = ascii_lowercase[:state_nb]
        self._model_free = "mf" in ext if ext else False
        self._temp_coef = "tc" in ext if ext else False

    @property
    def name(self) -> str:
        return self._name

    @property
    def states(self) -> str:
        return self._states

    @property
    def model_free(self) -> bool:
        return self._model_free

    @property
    def temp_coef(self) -> bool:
        return self._temp_coef


model = _Model()

set_model = model.set_model



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np
from scipy.optimize import root

from chemex.models.constraints import pop_2st
from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting
from chemex.parameters.userfunctions import user_function_registry

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions
    from chemex.typing import ArrayFloat

NAME = "2st_monomer_trimer"

TP = ("temperature", "p_total")


def calculate_residuals(
    concentrations: ArrayFloat,
    p_total: float,
    kd: float,
) -> ArrayFloat:
    monomer, trimer = concentrations
    return np.array(
        [
            monomer + 3.0 * trimer - p_total,
            kd * trimer - monomer**3,
        ],
    )


@lru_cache(maxsize=100)
def calculate_concentrations(p_total: float, kd: float) -> dict[str, float]:
    concentrations_start = (p_total, 0.0)
    results = root(calculate_residuals, concentrations_start, args=(p_total, kd))
    return {"monomer": results["x"][0], "trimer": results["x"][1]}


def make_settings_2st_monomer_trimer(
    conditions: Conditions,
) -> dict[str, ParamLocalSetting]:
    p_total = conditions.p_total
    if p_total is None:
        msg = f"'p_total' must be specified to use the '{NAME}' model"
        raise ValueError(msg)
    return {
        "koff": ParamLocalSetting(
            name_setting=NameSetting("koff", "", ("temperature",)),
            value=100.0,
            min=0.0,
            vary=True,
        ),
        "kd": ParamLocalSetting(
            name_setting=NameSetting("kd", "", ("temperature",)),
            value=1e-6,
            min=0.0,
            vary=True,
        ),
        "kon": ParamLocalSetting(
            name_setting=NameSetting("kon", "", ("temperature",)),
            min=0.0,
            expr="{koff} / max({kd}, 1e-32)",
        ),
        "c_monomer": ParamLocalSetting(
            name_setting=NameSetting("c_monomer", "", TP),
            expr=f"concentrations({p_total}, {{kd}})['monomer']",
        ),
        "c_trimer": ParamLocalSetting(
            name_setting=NameSetting("c_trimer", "", TP),
            expr=f"concentrations({p_total}, {{kd}})['trimer']",
        ),
        "kab": ParamLocalSetting(
            name_setting=NameSetting("kab", "", TP),
            expr="3.0 * {kon} * {c_monomer}",
        ),
        "kba": ParamLocalSetting(
            name_setting=NameSetting("kba", "", TP),
            expr="{koff}",
        ),
        "pa": ParamLocalSetting(
            name_setting=NameSetting("pa", "", TP),
            expr="pop_2st({kab}, {kba})['pa']",
        ),
        "pb": ParamLocalSetting(
            name_setting=NameSetting("pb", "", TP),
            expr="pop_2st({kab}, {kba})['pb']",
        ),
    }


def register() -> None:
    model_factory.register(name=NAME, setting_maker=make_settings_2st_monomer_trimer)
    user_functions = {
        "concentrations": calculate_concentrations,
        "pop_2st": pop_2st,
    }
    user_function_registry.register(name=NAME, user_functions=user_functions)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, Literal, Protocol, TypeVar

if TYPE_CHECKING:
    from chemex.containers.data import Data
    from chemex.nmr.spectrometer import Spectrometer


class NoFilterer:
    """Filterer that does no filtering."""

    def __init__(self, **_extra: Any) -> None:
        return

    def filter(self, _data: Data) -> None:
        return


class PlanesDataSettings(Protocol):
    filter_planes: list[int]


class PlanesExperimentConfig(Protocol):
    data: PlanesDataSettings


def _filter_planes(data: Data, planes_to_filter: list[int]) -> None:
    selection = [i for i in planes_to_filter if 0 <= i < data.size]
    data.mask[selection] = False


class PlanesFilterer:
    def __init__(self, config: PlanesExperimentConfig, **_extra: Any) -> None:
        self.config = config

    def filter(self, data: Data) -> None:
        _filter_planes(data, self.config.data.filter_planes)


class CestExperimentSettings(Protocol):
    observed_state: Literal["a", "b", "c", "d"]
    sw: float


class CestDataSettings(Protocol):
    filter_planes: list[int]
    filter_offsets: list[tuple[float, float]]
    filter_ref_planes: bool = False


class CestExperimentConfig(Protocol):
    experiment: CestExperimentSettings
    data: CestDataSettings


T = TypeVar("T", bound=CestExperimentConfig)


def _filter_ref_planes(data: Data) -> None:
    data.mask[data.refs] = False


def _filter_offsets(
    data: Data,
    config: CestExperimentConfig,
    spectrometer: Spectrometer,
) -> None:
    state = config.experiment.observed_state
    state_ppm = spectrometer.par_values[f"cs_i_{state}"]
    state_offset = spectrometer.ppms_to_offsets(state_ppm)

    offsets = data.metadata
    offset_min = min(offsets[~data.refs])

    for offset_to_filter, filter_bandwidth in config.data.filter_offsets:
        absolute_offset = offset_to_filter + state_offset
        offset_with_aliasing = (
            absolute_offset - offset_min
        ) % config.experiment.sw + offset_min
        shifted_offsets = offsets - offset_with_aliasing
        mask_filter = abs(shifted_offsets) < filter_bandwidth * 0.5
        data.mask[mask_filter] = False


class CestFilterer(Generic[T]):
    def __init__(self, config: T, spectrometer: Spectrometer, **_extra: Any) -> None:
        self.config = config
        self.spectrometer = spectrometer

    def filter(self, data: Data):
        _filter_planes(data, self.config.data.filter_planes)

        if self.config.data.filter_ref_planes:
            _filter_ref_planes(data)

        _filter_offsets(data, self.config, self.spectrometer)



"""The atom module defines the Atom class, used to identify atoms."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from .constants import CORRECT_ATOM_NAME
from .nucleus import Nucleus, str2nucleus

if TYPE_CHECKING:
    from collections.abc import Hashable


@dataclass(order=True)
class Atom:
    """Represents an atom in a molecular structure.

    This class encapsulates details about an atom, including its name,
    associated nucleus type, and search keys for matching and identification.

    Attributes:
        name (str): The name of the atom.
        nucleus (Nucleus): The type of nucleus associated with the atom, determined automatically.
        search_keys (set[Hashable]): Set of keys used for searching or matching the atom.
    """

    name: str
    nucleus: Nucleus = field(init=False, compare=False)
    search_keys: set[Hashable] = field(init=False, default_factory=set, compare=False)

    def __post_init__(self) -> None:
        """Initializes the Atom instance, correcting the atom name and determining the nucleus type."""
        name = self.name.strip().upper()
        self.name = CORRECT_ATOM_NAME.get(name, name)
        self.nucleus = str2nucleus(self.name[:1])
        self.search_keys.update({self, self.nucleus})

    def match(self, other: Atom) -> bool:
        """Checks if this atom matches another atom based on the name.

        Args:
            other (Atom): The other atom to compare with.

        Returns:
            bool: True if the other atom's name starts with the name of this atom, False otherwise.
        """
        return other.name.startswith(self.name)

    def __hash__(self) -> int:
        """Generates a hash value for an Atom instance.

        Returns:
            int: The hash value of the atom.
        """
        return hash(self.name)

    def __bool__(self) -> bool:
        """Determines the boolean value of the Atom instance.

        Returns:
            bool: True if the atom has a valid name, False otherwise.
        """
        return bool(self.name)

    def __str__(self) -> str:
        """Returns the string representation of the Atom instance.

        Returns:
            str: The name of the atom.
        """
        return self.name
